# -*- coding: utf-8 -*-

# Copyright © 2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
from .base import Delaunay, Voronoi
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from collections import OrderedDict
from warnings import warn


class KohonenMesh(Voronoi):
    r"""Kohonen fitting for arbitrary initial mesh.

    Attributes:
        step_size (tuple of float):
            :math:`(\epsilon_0, \epsilon_T, \gamma)` that determine step size
            :math:`\epsilon_t = a(b+t)^{-\gamma}` at iteration :math:`t`
            with :math:`T` the size of the dataset given to given to :meth:`tessellate`;
            default is ``(.1, .01, .55)``
        neighbourhood_weight (pair of float):
            weight for the winning node (first element) and its neighbours (second element);
            default is ``(1., .1)``
        pass_count (int or float):
            number of passes over the dataset; default is 5
        initial_mesh (Delaunay):
            mesh which cell centers are used as initial nodes;
            default consists of estimating the location density and randomly sampling
            the cell centers from this distribution
        initial_mesh_tessellate_kwargs (dict):
            arguments to :meth:`initial_mesh.tessellate`

    Trailing keyword arguments are passed to :meth:`initial_mesh.__init__`.
    The default mesh initialization expects arguments: *avg_probability*, *min_distance* and
    *max_distance*.

    """

    __slots__ = ('_step_size', '_neighbourhood_weight', 'pass_count',
        'initial_mesh', 'initial_mesh_init_kwargs', 'initial_mesh_tessellate_kwargs')

    def __init__(self, scaler=None, step_size=None, neighbourhood_weight=None, pass_count=None,
        initial_mesh=None, initial_mesh_tessellate_kwargs={}, **initial_mesh_init_kwargs):
        Voronoi.__init__(self, scaler)
        if step_size is None:
            step_size = (.1, .01, .55)
        self._step_size = step_size
        if neighbourhood_weight is None:
            neighbourhood_weight = (1., .1)
        self._neighbourhood_weight = neighbourhood_weight
        if pass_count is None:
            pass_count = 5
        self.pass_count = pass_count
        self.initial_mesh = initial_mesh
        self.initial_mesh_init_kwargs = initial_mesh_init_kwargs
        self.initial_mesh_tessellate_kwargs = initial_mesh_tessellate_kwargs

    def tessellate(self, points, **kwargs):
        # build initial mesh
        if self.initial_mesh is None:
            #from tramway.tessellation.grid import RegularMesh
            #self.initial_mesh = RegularMesh
            from tramway.tessellation.gwr.gas import Gas
            avg_probability = self.initial_mesh_init_kwargs['avg_probability']
            knn = 10#round(avg_probability * points.shape[0])
            rmin = self.initial_mesh_init_kwargs['min_distance'] * .01
            rmax = self.initial_mesh_init_kwargs['max_distance']
            if rmax is None:
                rmax = self.initial_mesh_init_kwargs['avg_distance'] * 2.
            X = np.asarray(self._preprocess(points))
            #if 10000 < X.shape[0]:
            #    X = X[np.random.randint(0, X.shape[0], 10000)]
            radius = Gas(X).boxed_radius(X, knn, rmin, rmax)
            density = 1. / radius
            density /= np.sum(density)
            P = np.r_[0, np.cumsum(density)]
            p = np.sort(np.random.rand(int(round(1./avg_probability))))
            i, j = [], 0
            for q in p:
                while P[j] <= q:
                    j += 1
                i.append(j-1)
            #i = [ ((P[:-1] <= _p) & (_p < P[1:])).nonzero()[0][0].tolist() for _p in p ]
            self.initial_mesh = Voronoi(self.scaler)
            if isinstance(points, pd.DataFrame):
                X = points.loc[i]
            else:
                X = points[i]
            self.initial_mesh.tessellate(X)
        if callable(self.initial_mesh):
            self.initial_mesh = self.initial_mesh(self.scaler, **self.initial_mesh_init_kwargs)
            self.initial_mesh.tessellate(points, **self.initial_mesh_tessellate_kwargs)
            ## copy a few pointers
            self._cell_centers = self.initial_mesh._cell_centers
            self._cell_label = self.initial_mesh.cell_label
            #self._cell_adjacency = self.initial_mesh.cell_adjacency
            #self._adjacency_label = self.initial_mesh.adjacency_label
        elif isinstance(self.initial_mesh, Delaunay):
            ## copy a few pointers
            self._cell_centers = self.initial_mesh._cell_centers
            self._cell_label = self.initial_mesh.cell_label
            #self._cell_adjacency = self.initial_mesh.cell_adjacency
            #self._adjacency_label = self.initial_mesh.adjacency_label
            pass
        else:
            raise NotImplementedError
        # step size
        T = points.shape[0]
        eps_0, eps_T, gamma = self._step_size
        c = (eps_T / eps_0) ** (-1. / gamma)
        b = float(T) / (c - 1.)
        a = eps_0 / (b ** (-gamma))
        self._step_size = (a, b, gamma)
        # learn
        X = np.asarray(self.initial_mesh._preprocess(points))
        W = self._cell_centers
        t_max = T * self.pass_count
        t = 0
        while t < t_max:
            i = np.random.randint(T)
            x = X[i]
            dW = x - W
            u = np.argmin(np.sum(dW * dW, axis=1))
            W[u] += self.neighbourhood_weight(0, t) * self.step_size(t) * dW[u]
            visited = U = set([u])
            for d in range(1, len(self._neighbourhood_weight)):
                V = set()
                for u in U:
                    V |= set(self.initial_mesh.neighbours(u).tolist())
                V -= visited
                if not V:
                    break
                V_ = list(V)
                W[V_] += self.neighbourhood_weight(d, t) * self.step_size(t) * dW[V_]
                visited |= V
                U = V
            t += 1

    def step_size(self, t):
        a, b, gamma = self._step_size
        return a * (b + float(t)) ** (-gamma)

    def neighbourhood_weight(self, d, t):
        return self._neighbourhood_weight[d]

    # cell_centers property
    @property
    def cell_centers(self):
        return self.initial_mesh.cell_centers

    @cell_centers.setter
    def cell_centers(self, centers):
        self.initial_mesh.cell_centers = centers
        self._cell_centers = self.initial_mesh._cell_centers

    # cell_label
    @property
    def cell_label(self):
        return self.initial_mesh.cell_label

    @cell_label.setter
    def cell_label(self, labels):
        self.initial_mesh.cell_label = labels

    # cell_adjacency property
    @property
    def cell_adjacency(self):
        return self.initial_mesh.cell_adjacency

    @cell_adjacency.setter
    def cell_adjacency(self, matrix):
        self.initial_mesh.cell_adjacency = matrix
        self._cell_adjacency = self.initial_mesh._cell_adjacency

    # adjacency label
    @property
    def adjacency_label(self):
        return self.initial_mesh.adjacency_label

    @adjacency_label.setter
    def adjacency_label(self, labels):
        self.initial_mesh.adjacency_label = labels
        self._adjacency_label = self.initial_mesh._adjacency_label


setup = {
    'make_arguments': OrderedDict((
        ('min_distance', ()),
        ('avg_distance', ()),
        ('max_distance', ()),
        ('avg_probability', ()),
        # avg_location_count allows to control avg_probability from the commandline
        ('avg_location_count', dict(args=('-c', '--location-count'), kwargs=dict(type=int, default=80, help='average number of locations per cell'), translate=True)),
        )),
    }


__all__ = ['setup', 'KohonenMesh']

import sys
if sys.version_info[0] < 3:

    import rwa
    import tramway.core.hdf5 as rules
    kohonen_mesh_exposes = rules.voronoi_exposes + list(KohonenMesh.__slots__)
    rwa.hdf5_storable(
        rwa.default_storable(KohonenMesh, exposes=kohonen_mesh_exposes),
        agnostic=True)

    __all__.append('kohonen_mesh_exposes')




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
.. =============================================
.. Alternative implementation for :class:`Graph`
.. =============================================

"""

from ..graph import *
import numpy as np
import scipy.sparse as sp
from collections import deque
import itertools
import time


class ArrayGraph(Graph):
    """With DictGraph backend for Gas::

        %run glycine_receptor.py
        Elapsed:  mean: 40106 ms  std: 5819 ms

    With ArrayGraph backend::

        %run glycine_receptor.py
        Elapsed:  mean: 2937 ms  std: 304 ms
        ratio: 13.6554

    """
    __slots__ = ('node_capacity', 'edge_capacity', '_node_counter', \
        '_edge_counter', '_free_nodes', '_free_edges', 'nodes', 'edges', 'adjacency', \
        '_fast_node')

    def __init__(self, node_defaults, edge_defaults, node_count=None, edge_count=None):
        Graph.__init__(self, node_defaults, edge_defaults)
        if node_count:
            if not edge_count:
                edge_count = node_count * 4
        elif edge_count:
            node_count = edge_count / 4
        else:
            node_count = 10000
            edge_count = node_count * 4
        self.node_capacity = node_count
        self.edge_capacity = edge_count
        self._node_counter = 0
        self._edge_counter = 0
        self._free_nodes = deque()
        self._free_edges = deque()
        self.nodes = {}
        self.edges = {}
        self.adjacency = sp.lil_matrix((self.node_capacity, self.node_capacity), \
            dtype=int)
        self._fast_node = {}

    def assert_node(self, n):
        if not self.has_node(n):
            raise MissingNodeError(n)

    def assert_edge(self, e):
        if self._edge_misses(e):
            raise MissingEdgeError(e)

    def connect(self, n1, n2, **kwargs):
        self.assert_node(n1)
        self.assert_node(n2)
        if not self.edges:
            self._init_edges(**kwargs)
        e = self.adjacency[n1, n2] - 1
        if e < 0: # e is 0 does not work as expected because of types
            if self._free_edges:
                e = self._free_edges.pop()
            else:
                e = self._edge_counter
                self._edge_counter += 1
                while self.edge_capacity <= self._edge_counter: # actual index is e - 1
                    self._resize_edges()
            try:
                self.adjacency[n1, n2] = e + 1
            except IndexError:
                print((self.adjacency.rows[n1], self.adjacency.data[n1]))
                raise
            self.adjacency[n2, n1] = e + 1
        for attr, default in self.edge_defaults.items():
            self.edges[attr][e] = kwargs.get(attr, default)

    def unsafe_disconnect(self, n1, n2):
        edge = self.adjacency[n1, n2] - 1
        self.adjacency[n1, n2] = 0
        self.adjacency[n2, n1] = 0
        self._free_edges.append(edge)

    def disconnect(self, n1, n2, edge=None):
        if not isinstance(n2, list):
            n2 = [n2]
        if n2:
            #print(('disconnect',(n1, type(n1)),(n2, type(n2))))
            self.assert_node(n1)
            for n in n2:
                self.assert_node(n)
                self.unsafe_disconnect(n1, n)

    def find_edge(self, n1, n2):
        e = self.adjacency[n1, n2]
        if e:   return e - 1
        else:   return None

    def get_node_attr(self, n, attr):
        #print(('get_node_attr', n, attr))
        self.assert_node(n)
        try:
            return self.nodes[attr][n]
        except KeyError:
            raise NodeAttributeError(attr)

    def set_node_attr(self, n, **kwargs):
        self.assert_node(n)
        for attr, val in kwargs.items():
            if attr in self.nodes:
                self.nodes[attr][n] = val
                if attr in self._fast_node:
                    self._fast_node[attr][n] = np.dot(val, val)
            else: raise NodeAttributeError(attr)

    def get_edge_attr(self, e, attr):
        self.assert_edge(e)
        try:
            return self.edges[attr][e]
        except KeyError:
            raise EdgeAttributeError(attr)

    def set_edge_attr(self, e, **kwargs):
        self.assert_edge(e)
        for attr, val in kwargs.items():
            if attr in self.edges:
                self.edges[attr][e] = val
            else: raise EdgeAttributeError(attr)

    def iter_edges(self):
        """Potentially inefficient."""
        return [e - 1 for e in set(itertools.chain(self.adjacency.data))]

    def iter_edges_from(self, n):
        self.assert_node(n)
        return zip([e - 1 for e in self.adjacency.data[n]], self.adjacency.rows[n])

    def iter_neighbors(self, n):
        self.assert_node(n)
        return self.adjacency.rows[n]

    def has_node(self, n):
        return n < self._node_counter and n not in self._free_nodes

    def _edge_misses(self, e):
        # _edge_counter is incremented before assignement, i.e. edges[*] has size at least
        # _edge_counter+1
        return self._edge_counter <= e or e in self._free_edges

    def add_node(self, **kwargs):
        if not self.nodes:
            self._init_nodes(**kwargs)
        if self._free_nodes:
            n = self._free_nodes.pop()
            assert n < self._node_counter
        else:
            while self.node_capacity <= self._node_counter:
                self._resize_nodes()
            n = self._node_counter
            self._node_counter += 1
        for attr, default in self.node_defaults.items():
            self.nodes[attr][n] = kwargs.get(attr, default)
        for attr in self._fast_node:
            w = self.nodes[attr][n]
            self._fast_node[attr][n] = np.dot(w, w)
        return n

    def del_node(self, n):
        #print(('del_node', n, type(n)))
        self.assert_node(n)
        neighbors = list(self.iter_neighbors(n))
        if neighbors:
            self.disconnect(n, neighbors)
        self._free_nodes.append(n)

    @property
    def size(self):
        return self._node_counter - len(self._free_nodes)

    def iter_nodes(self):
        """Avoid by all means!"""
        free = set(self._free_nodes)
        return [ n for n in range(0, self._node_counter) if n not in free ]

    def _init_nodes(self, **kwargs):
        for attr, default in self.node_defaults.items():
            if default is None:
                if attr in kwargs:
                    val = kwargs[attr]
                else: raise NodeAttributeError(attr)
            else:   val = default
            if isinstance(val, np.ndarray):
                dim = val.size
                typ = val.dtype
            else:
                dim = 1
                typ = type(val)
            self.nodes[attr] = np.zeros((self.node_capacity, dim), dtype=typ)

    def _init_edges(self, **kwargs):
        for attr, default in self.edge_defaults.items():
            if default is None:
                if attr in kwargs:
                    val = kwargs[attr]
                else: raise NodeAttributeError(attr)
            else:   val = default
            if isinstance(val, np.ndarray):
                dim = val.size
                typ = val.dtype
            else:
                dim = 1
                typ = type(val)
            self.edges[attr] = np.zeros((self.edge_capacity, dim), dtype=typ)

    def _resize_nodes(self):
        for attr in self.nodes:
            self.nodes[attr] = np.r_[self.nodes[attr], np.zeros_like(self.nodes[attr])]
        for attr in self._fast_node:
            self._fast_node[attr] = np.concatenate((
                    self._fast_node[attr],
                    np.zeros_like(self._fast_node[attr])
                ))
        self.node_capacity *= 2
        # properly reshape lil matrix
        rows, data = self.adjacency.rows, self.adjacency.data
        self.adjacency = sp.lil_matrix((self.node_capacity, self.node_capacity),
            dtype=self.adjacency.dtype)
        self.adjacency.rows[:rows.size] = rows
        self.adjacency.data[:data.size] = data

    def _resize_edges(self):
        for attr in self.edges:
            self.edges[attr] = np.r_[self.edges[attr], np.zeros_like(self.edges[attr])]
        self.edge_capacity *= 2

    def export(self, sparse='csr'):
        #t = time.time()
        I = np.ones(self.node_capacity, dtype=bool)
        I[self._node_counter:] = 0
        I[self._free_nodes] = 0
        V = {}
        for attr in self.nodes:
            V[attr] = self.nodes[attr][I]
        #print('{:.0f}'.format((time.time() - t) * 1e6))
        #t = time.time()
        #Iadj, = np.nonzero(self.adjacency.rows)
        #V = {}
        #for attr in self.nodes:
        #       V[attr] = self.nodes[attr][Iadj]
        #print('{:.0f}'.format((time.time() - t) * 1e6))
        #assert np.all(I[Iadj]) and Iadj.size == np.sum(I)
        A = self.adjacency[I].tocsc()[:,I]
        if sparse == 'csc':
            A = A.tocsc()
        elif sparse == 'csr':
            A = A.tocsr()
        elif sparse == 'bsr':
            A = A.tobsr()
        elif sparse == 'coo':
            A = A.tocoo()
        else:
            raise NotImplementedError
        J = np.ones(self.edge_capacity, dtype=bool)
        J[self._edge_counter:] = 0
        J[self._free_edges] = 0
        E = {}
        for attr in self.edges:
            E[attr] = self.edges[attr][J]
        J, = np.nonzero(J)
        K = np.zeros(self._edge_counter + 1, dtype=int) # could be A.shape[0] * np.ones to
        # detect edge coding errors, but it seems there is none
        K[J] = np.arange(0, J.size)
        A.data = K[A.data - 1]
        #print((A.data.min(), A.data.max()))
        return (A, V, E)

    def square_distance(self, attr, eta, eta2=None):
        if eta2 is None:
            eta2 = np.dot(eta, eta)
        if attr not in self._fast_node:
            self._fast_node[attr] = np.zeros(self.nodes[attr].shape[0], \
                dtype=self.nodes[attr].dtype)
            w = self.nodes[attr][:self._node_counter]
            self._fast_node[attr][:self._node_counter] = np.sum(w * w, axis=1)
        d = self._fast_node[attr][:self._node_counter] + eta2 - 2.0 * \
            np.dot(self.nodes[attr][:self._node_counter], eta)
        d[self._free_nodes] = np.nan#float('nan')
        return (d, lambda i: i)




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import scipy.sparse.csgraph as csgraph
import itertools


def dilation(node, graph, step=1, boundary=None):
    """
    Expand across a graph from a node, or equivalently from a cell across an adjacency matrix.

    Graph expansions can be further expanded::

        crawled_nodes = dilation(node, graph, step=slice(3))
        more_crawled_nodes = dilation(crawled_nodes, graph, step=slice(8))

    Arguments:

        node (int or dict of sets): row/column index in `graph` (node/cell index)
            or partial `dilation` crawl.

        graph (scipy.sparse.csr_matrix): square adjacency matrix.

        step (int or (int, int)): minimum and maximum (+1) numbers of dilation steps.

        boundary (list or set): nodes to be excluded.

    Returns:

        dict of sets: sets of node indices in a :class:`dict` with steps as keys.
    """
    if boundary and not isinstance(boundary, set):
        boundary = set(boundary)
    if isinstance(step, slice):
        smin, smax = step.start, step.stop
    elif isinstance(step, tuple):
        smin, smax = step
    else:
        smin = step
        smax = smin + 1
    if np.isscalar(node):
        try:
            graph.indices[node]
        except IndexError:
            raise IndexError('node index out of range')
        except:
            raise TypeError('graph is not a csr sparse matrix')
        nodes = {0: {node}}
        s = 0
    elif isinstance(node, dict):
        nodes = node
        if not nodes:
            return nodes
        s = max(nodes.keys())
    if smax is not None and smax <= s + 1:
        # may raise KeyError; this is alright
        nodes = { _s: nodes[_s] for _s in range(smin, smax) }
        if boundary:
            s = smax - 1
            while nodes:
                nodes[s] -= boundary
                if nodes[s]:
                    break
                else:
                    del nodes[s]
                    s -= 1
            for _s in nodes:
                if nodes[_s] & boundary:
                    raise NotImplementedError
        return nodes
    while smax is None or s + 1 < smax:
        ns = set()
        for n in nodes[s]:
            ns |= set(graph.indices[graph.indptr[n]:graph.indptr[n+1]])
        for t in nodes:
            ns -= nodes[t]
        if boundary:
            ns -= boundary
        # stopping criteria:
        if not ns:
            break
        s += 1
        nodes[s] = ns
    if smin:
        if smax is not None and smin == smax - 1:
            try:
                nodes = {smin: nodes[smin]}
            except KeyError:
                nodes = {}
        else:
            nodes = { s: nodes[s] for s in nodes if smin <= s }
    return nodes


class NoSolutionError(ValueError):
    pass


def memoize(f):
    y_name = 'distance'
    _context = None
    known_computation = {}
    def wrapper(x, context, *args, **kwargs):
        if args:
            args = list(args)
            y = args.pop(0)
        else:
            try:
                y = kwargs.pop(y_name)
            except KeyError:
                return f(x, context, **kwargs) # do not cache
        print((known_computation, y_name))
        if _context is not context:
            _context = context
            known_computation = {} # new memoization table
        try:
            ret = known_computation[(x, y)]
        except KeyError:
            ret = known_computation[(x, y)] = f(x, context, y, *args, **kwargs)
        return ret
    return wrapper

#@memoize
def contour(cell, adjacency, distance=1, nodes=None, dilation_adjacency=None,
        fallback=False, debug=False, cells=None):
    # cells was useful to debugging
    if nodes is None:
        if dilation_adjacency is None:
            dilation_adjacency = adjacency
        nodes = dilation(cell, dilation_adjacency, step=slice(distance+1))
    try:
        ns = nodes[distance]
    except KeyError:
        if fallback:
            return ([], False)
        else:
            raise NoSolutionError
    strings = []
    while ns:
        seed = ns.pop()
        string = [seed]
        for _ in range(2):
            _any = seed
            string = string[::-1]
            while True:
                _next = set(adjacency.indices[adjacency.indptr[_any]:adjacency.indptr[_any+1]])
                _next &= ns
                if _next:
                    _any = _next.pop()
                    ns.remove(_any)
                    string.append(_any)
                else:
                    break
        if string[1:]: # discard isolated points
            strings.append(string)
    if not strings[1:]:
        string = strings[0]
        _next = adjacency.indices[adjacency.indptr[string[-1]]:adjacency.indptr[string[-1]+1]]
        if string[0] in _next:
            if fallback:
                return (string, True)
            else:
                return string
    # get the solution at distance-1
    if distance == 1:
        if fallback:
            return ([], False)
        else:
            raise NoSolutionError
    # inner contour
    cont = contour(cell, adjacency, distance=distance-1, nodes=nodes,
        dilation_adjacency=dilation_adjacency, fallback=fallback, debug=debug, cells=cells)
    if fallback:
        if cont[1]:
            cont = cont[0]
        else:
            return cont
    inner = set(cont)
    min_length = len(inner)
    any_inserted = False
    # insert one string at a time into the inner contour
    for graft in strings:
        # first shorten `graft` so that both ends stand near `cont`
        try:
            _graft_bracket = []
            for _i, _di in ((0,1),(-1,-1)):
                while True:
                    graft[_i+_di] # raise IndexError if no such element
                    _head = graft[_i]
                    _neighbours = set(adjacency.indices[adjacency.indptr[_head]:adjacency.indptr[_head+1]]) & inner
                    if _neighbours:
                        _graft_bracket.append(_i)
                        break
                    _i += _di
            _i, _j = _graft_bracket
            if _j < -1:
                graft = graft[_i:_j+1]
            elif 0 < _i:
                graft = graft[_i:]
            if not graft[1:]: # if len(graft) < 2
                raise IndexError
        except (IndexError, ValueError):
            if debug:
                print('skipping graft: no contact point')
            continue
        # find the extreme (bracket) inner nodes
        outer = set(graft)
        _all_neighbours = [ set(adjacency.indices[adjacency.indptr[_n]:adjacency.indptr[_n+1]])
                for _n in graft ]
        _inner_neighbours = set()
        for _a in _all_neighbours:
            _inner_neighbours |= _a
        _inner_neighbours &= inner
        _inner_targets = _bracket_elements(_inner_neighbours, adjacency, (graft[0], graft[-1]),
            cont, len(graft), debug=debug)
        if len(_inner_targets) != 2:
            if debug:
                print('skipping graft: elements are not contiguous')
            continue
        # find the contact nodes, inner and outer
        join = []
        for _target in _inner_targets:
            _outer_neighbours = []
            for _n, _a in zip(graft, _all_neighbours):
                if _target in _a:
                    _outer_neighbours.append((_n, _a))
            _subsets = [outer, inner]
            while _outer_neighbours[1:] and _subsets[1:]:
                # while multiple solutions
                # first favor the lesser number of inner neighbours
                # and then favor the lesser number of outer neighbours
                _subset = _subsets.pop()
                _min_count = len(_subset) + 1
                for _n, _a in _outer_neighbours:
                    _c = len(_a & _subset)
                    if _c < _min_count:
                        _candidates = []
                        _min_count = _c
                    if _c == _min_count:
                        _candidates.append((_n, _a))
                _outer_neighbours = _candidates
            join.append((_target, _outer_neighbours[0][0]))
        # get the indices of the contact nodes
        _inner2, _outer2 = join.pop()
        _inner1, _outer1 = join.pop()
        _remove = _inner_neighbours - set((_inner1, _inner2))
        _i1, _i2 = cont.index(_inner1), cont.index(_inner2)
        _j1, _j2 = graft.index(_outer1), graft.index(_outer2)
        _dj = 1 if _j1 <= _j2 else -1
        # assertions
        if debug:
            _ni, _nj = cont[_i1], graft[_j1]
            assert _nj in adjacency.indices[adjacency.indptr[_ni]:adjacency.indptr[_ni+1]]
            _ni, _nj = cont[_i2], graft[_j2]
            assert _nj in adjacency.indices[adjacency.indptr[_ni]:adjacency.indptr[_ni+1]]
        # solution 1: cut `cont` from _i1 to _i2
        # solution 2: cut `cont` from _i2 to _i1
        graft_length = abs(_j2 - _j1) + 1
        if graft_length < 2:
            if debug:
                print('skipping graft: too short')
            continue
        if _i1 < _i2:
            len1 = graft_length + len(cont) - (_i2 - _i1 - 1)
            len2 = graft_length + (_i2 - _i1 + 1)
            cut1 = set(cont[_i1+1:_i2]) # solution 1: cont[:_i1+1]+graft+cont[_i2:]
            cut2 = set(cont[:_i1] + cont[_i2+1:]) # solution 2: graft + cont[_i1:_i2+1]
        else:
            len1 = graft_length + (_i1 - _i2 + 1)
            len2 = graft_length + len(cont) - (_i1 - _i2 - 1)
            cut1 = set(cont[:_i2] + cont[_i1+1:])
            cut2 = set(cont[_i2+1:_i1])
        count1, count2 = [ len(_cut & _remove) for _cut in (cut1, cut2) ]
        if count1 < count2 and min_length <= len2:
            # solution 2 wins
            _dj = -_dj
            _j1 += _dj
            if _j1 == -1:
                _j1 = None
            graft = graft[_j2:_j1:_dj]
            if _i1 < _i2:
                cont = graft + cont[_i1:_i2+1]
            else:
                cont = cont[:_i2+1] + graft + cont[_i1:]
            assert len(cont) == len2
        elif count2 < count1 and min_length <= len1:
            # solution 1 wins
            _j2 += _dj
            if _j2 == -1:
                _j2 = None
            graft = graft[_j1:_j2:_dj]
            if _i1 < _i2:
                cont = cont[:_i1+1] + graft + cont[_i2:]
            else:
                cont = graft + cont[_i2:_i1+1]
            assert len(cont) == len1
        else:
            if debug:
                print('skipping graft: conflict between solutions')
                print((count1, len1, count2, len2))
                _diagnose((cont, graft, [_inner1, _outer1], [_inner2, _outer2]), cells)
            continue
        any_inserted = True
    if fallback:
        return (cont, any_inserted)
    else:
        if any_inserted:
            return cont
        else:
            raise NoSolutionError


def _bracket_elements(string, adjacency, beacon=None, fullstring=None, outerlen=None, debug=False):
    if 2 < len(string):
        bracket = set()
        for n in string:
            neighbours = set(adjacency.indices[adjacency.indptr[n]:adjacency.indptr[n+1]])
            neighbours &= string
            if len(neighbours) == 1:
                bracket.add(n)
        if beacon and 2 < len(bracket):
            _bracket = set()
            for ns in beacon:
                if not isinstance(ns, (tuple, list, set)):
                    ns = [ns]
                _ns = set(bracket) # copy
                for n in ns:
                    _ns &= set(adjacency.indices[adjacency.indptr[n]:adjacency.indptr[n+1]])
                try:
                    _bracket.add(_ns.pop())
                except KeyError:
                    if debug:
                        print('bracketing: orphan beacon')
                    pass
                if _ns and debug:
                    print('bracketing: choosing between nodes at random')
            bracket = _bracket
    else:
        bracket = set(string) # copy
    return bracket


def _diagnose(paths, cells):
    if cells is None:
        return
    import matplotlib.pyplot as plt
    import tramway.plot.mesh as mesh
    f = plt.figure()
    mesh.plot_delaunay(cells, axes=f.gca())
    colors = 'bgkyr'
    for p, c in zip(paths, colors):
        x = cells.tesselation.cell_centers[p]
        plt.plot(x[:,0], x[:,1], c, linestyle='-', linewidth=3)
    plt.show()




# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov


import logging
import math
import unittest
from multiprocessing import freeze_support

import numpy as np
from scipy.integrate import dblquad, quad

from .calculate_bayes_factors import calculate_bayes_factors
from .calculate_marginalized_integral import calculate_marginalized_integral
from .calculate_posteriors import (calculate_one_1D_posterior_in_2D,
                                   calculate_one_1D_prior_in_2D,
                                   calculate_one_2D_posterior)
from .convenience_functions import n_pi_func, p


class bayes_test(unittest.TestCase):
    """Tests for Bayes factor calculations"""

    def setUp(self):
        """Initialize data for different tests"""
        self.full_tests = False
        self.tol = 1e-32
        self.rel_tol = 1e-2
        self.B_threshold = 10.0
        self.t1_zeta_sps = np.asarray([[1.0, 1.0], [2.0, 3.0]])
        self.t1_zeta_ts = np.asarray([[2.0, 1.54], [-4.0, 0.0]])
        self.t1_ns = np.asarray([10, 5])
        self.t1_Vs = np.asarray([1, 0.2])
        self.t1_Vs_pi = np.asarray([0.3, 0.5])
        self.t1_us = np.divide(self.t1_Vs_pi, self.t1_Vs)

    def test_marginalized_integral(self):
        # Load data
        dim = 2
        n_pi = n_pi_func(dim)

        # >> Test case E = 0 <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [-0.3, -0.2]
        n = 20
        u = 0.95
        rel_loc_error = 0.77
        eta = np.sqrt(n_pi / (n + n_pi))
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=0.0, rel_loc_error=rel_loc_error)
        true_res = 3.496167136E-23
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed for E = 0")

        # >> Check result with 0 break points <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [-0.3, -0.2]
        n = 20
        u = 0.95
        rel_loc_error = 0.77
        eta = np.sqrt(n_pi / (n + n_pi))
        # pow = dim * (n + n_pi + 1.0) / 2.0 - 2.0
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        # print([n_pi, n, u, v0])
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=eta ** 2.0, rel_loc_error=rel_loc_error)
        true_res = 3.182820939E-23
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed for 0 break points. The obtained value %.2g does not match the expected %.2g" % (res, true_res))

        # >> Check result with 1 break points <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [-0.3, 0.6]
        n = 20
        u = 0.95
        rel_loc_error = 0.42
        eta = np.sqrt(n_pi / (n + n_pi))
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=eta ** 2.0, rel_loc_error=rel_loc_error)
        true_res = 1.457477573E-28
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed for 1 break points. The obtained value %.2g does not match the expected %.2g" % (res, true_res))

        # >> Check result with 2 break points <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [0.8, 0.6]
        n = 20
        u = 0.95
        rel_loc_error = 2.30
        eta = np.sqrt(n_pi / (n + n_pi))
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=eta ** 2.0, rel_loc_error=rel_loc_error)
        true_res = 6.019713659E-14
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed for 2 break points. The obtained value %.8g does not match the expected %.8g" % (res, true_res))

        # >> Check the result with zeta_sp = 0 <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [0.0, 0.0]
        n = 20
        u = 0.95
        rel_loc_error = 1.47
        eta = np.sqrt(n_pi / (n + n_pi))
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=eta ** 2.0, rel_loc_error=rel_loc_error)
        true_res = 1.145033778E-17
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed for zeta_sp = 0. The obtained value %.8g does not match the expected %.8g" % (res, true_res))

        # >> Check the result with zeta_a <<
        zeta_t = [0.7, 0.4]
        zeta_sp = [0.8, 0.6]
        zeta_a = [-0.1, -0.1]
        n = 20
        u = 0.95
        rel_loc_error = 1.47
        eta = np.sqrt(n_pi / (n + n_pi))
        pow = p(n, dim)
        v0 = 1 + n_pi / n * u
        res = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                              p=pow, v=v0, E=eta ** 2.0, rel_loc_error=rel_loc_error, zeta_a=zeta_a)
        true_res = 1.225618999E-17
        self.assertTrue(np.isclose(res, true_res, rtol=self.rel_tol, atol=self.tol),
                        "Marginalized integral calculation test failed with zeta_a. The obtained value %.8g does not match the expected %.8g" % (res, true_res))

    def test_bayes_factors(self):

        # >> Test one input bin <<
        zeta_ts = np.asarray([[0.7, 0.4]])
        zeta_sps = np.asarray([[0.8, 0.6]])
        ns = np.asarray([[20]])
        Vs = np.asarray([[0.8 ** 2.0]])
        us = np.asarray([[0.95]])
        loc_error = 0.1**2.0
        Vs_pi = us * Vs
        lg_Bs, forces, _ = calculate_bayes_factors(
            zeta_ts=zeta_ts, zeta_sps=zeta_sps, ns=ns, Vs=Vs, Vs_pi=Vs_pi, loc_error=loc_error)
        true_B = 0.2974282533

        # Check value
        self.assertTrue(np.isclose(10**lg_Bs[0], true_B, rtol=self.rel_tol, atol=self.tol),
                        "Bayes factor calculation failed for one bin. The obtained B = %.8g does not match the expected B = %.8g" % (10**lg_Bs[0, 0], true_B))
        # Check force presence
        self.assertTrue((true_B >= self.B_threshold) ==
                        forces[0], "Boolean conservative force return incorrect for the case of one bin")

        # >> Test zero localization error <<
        zeta_ts = np.asarray([[0.7, 0.4]])
        zeta_sps = np.asarray([[0.8, 0.6]])
        ns = np.asarray([[20]])
        Vs = np.asarray([[0.8 ** 2.0]])
        us = np.asarray([[0.95]])
        loc_error = 0
        Vs_pi = us * Vs
        lg_Bs, forces, _ = calculate_bayes_factors(
            zeta_ts=zeta_ts, zeta_sps=zeta_sps, ns=ns, Vs=Vs, Vs_pi=Vs_pi, loc_error=loc_error)
        true_B = 0.2974282533

        # Check value
        self.assertTrue(np.isclose(10**lg_Bs[0], true_B, rtol=self.rel_tol, atol=self.tol),
                        "Bayes factor calculation failed for one bin. The obtained B = %.8g does not match the expected B = %.8g" % (10**lg_Bs[0, 0], true_B))
        # Check force presence
        self.assertTrue((true_B >= self.B_threshold) ==
                        forces[0], "Boolean conservative force return incorrect for the case of one bin")

        # >> Test three input bins <<
        zeta_ts = np.asarray([[0.4, 0.3], [-1.0, 2.3], [-1.1, -0.33]])
        zeta_sps = np.asarray([[0.8, 0.6], [-0.45, 0.67], [6.32, 0.115]])
        ns = np.asarray([[500, 100, 3]])
        Vs = np.asarray([[0.4 ** 2.0, 1.25 ** 2.0, 0.3 ** 2.0]])
        us = np.asarray([[0.95, 8.7, 2.45]])
        loc_error = 0.1**2.0
        ns = ns.T
        Vs = Vs.T
        us = us.T
        Vs_pi = us * Vs
        N = len(ns)
        lg_Bs, forces, _ = calculate_bayes_factors(
            zeta_ts=zeta_ts, zeta_sps=zeta_sps, ns=ns, Vs=Vs, Vs_pi=Vs_pi, loc_error=loc_error)
        true_Bs = [0.05997370802, 3.666049402e49, 1.378104868]
        # print(Bs)
        # print (forces)

        for i in range(N):
            # Check value
            self.assertTrue(np.isclose(10**lg_Bs[i], true_Bs[i], rtol=self.rel_tol, atol=self.tol),
                            "Bayes factor calculation failed for one bin. For bin no. %i, the obtained B = %.8g does not match the expected B = %.8g" % (i + 1, 10**lg_Bs[i], true_Bs[i]))
            # Check force presence
            true_forces = (1 * (np.log10(true_Bs[i]) >= np.log10(self.B_threshold)) -
                           1 * (np.log10(true_Bs[i]) <= -np.log10(self.B_threshold)))
            self.assertTrue(true_forces == forces[i],
                            "For bin %i, the boolean force prediction (%r) did not correspond to expected value (%r)" % (i, true_Bs[i] >= self.B_threshold, forces[i]))

    def test_minimal_n(self):
        """
        Test calculations of the minimal number of jumps per bin to produce strong evidence.
        """
        # Load data
        B_threshold = 10
        # >> Test three input bins <<
        zeta_ts = np.asarray([[0.7, 0.5], [-1.0, 2.3], [-1.1, -0.33]])
        zeta_sps = np.asarray([[0.8, 0.6], [-0.45, 0.67], [6.32, 0.115]])
        ns = np.asarray([[20, 100, 3]])
        Vs = np.asarray([[0.8 ** 2.0, 1.25 ** 2.0, 0.3 ** 2.0]])
        us = np.asarray([[0.95, 8.7, 2.45]])
        loc_error = 0.2 ** 2.0
        ns = ns.T
        Vs = Vs.T
        us = us.T
        Vs_pi = us * Vs
        # N = len(ns)

        # Calculate min_ns
        lg_Bs, _, min_ns = calculate_bayes_factors(
            zeta_ts=zeta_ts, zeta_sps=zeta_sps, ns=ns, Vs=Vs, Vs_pi=Vs_pi, loc_error=loc_error)
        # Calculate Bs for min_ns to check
        lg_Bs, _, _ = calculate_bayes_factors(
            zeta_ts=zeta_ts, zeta_sps=zeta_sps, ns=min_ns, Vs=Vs, Vs_pi=Vs_pi, loc_error=loc_error)

        # Perform check
        self.assertTrue(np.all(np.abs(lg_Bs[0]) >= np.log10(B_threshold)),
                        "Not all of the Bayes factors for the minimal ns returned strong evidence. lg_Bs: %s" % (lg_Bs))
        # print(Bs)

    def test_posteriors_and_priors(self):

        # >> Test that the 2D zeta_a posterior is normalized <<
        zeta_t = np.asarray([0.7, 0.4])
        zeta_sp = np.asarray([0.8, 0.6])
        n = np.asarray(20)
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        loc_error = 0.1**2.0
        V_pi = u * V
        lamb = 'marg'
        tol = 1e-4

        # Wrapper to get just a function of zeta_ax, zeta_ay
        def integrate_me(zay, zax):
            return calculate_one_2D_posterior(zeta_t=zeta_t, zeta_sp=zeta_sp, zeta_a=[zax, zay], n=n, V=V, V_pi=V_pi, loc_error=loc_error, dim=2, lamb=lamb)

        lims = np.array([-1, 1]) * 10

        if self.full_tests:
            norm = dblquad(integrate_me, lims[0], lims[1],
                           lims[0], lims[1], epsabs=tol, epsrel=tol)[0]
            true_norm = 1

            # Check value
            self.assertTrue(np.isclose(norm, true_norm, rtol=self.rel_tol, atol=self.tol),
                            "Active force posterior normalization test failed. Obtained norm = %.8g did not match the expected norm = %.8g" % (norm, true_norm))

        # >> Test that 1D zeta_ax posterior is normalized in 2D with localization error <<
        zeta_t = np.asarray([0.7, 0.4])
        zeta_sp = np.asarray([0.8, 0.6])
        n = np.asarray(20)
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        loc_error = 0.1**2.0
        V_pi = u * V
        lamb = 0  # 'marg'
        tol = 1e-4

        # Wrapper to get just a function of zeta_ax, zeta_ay
        def integrate_me(zax):
            return calculate_one_1D_posterior_in_2D(zeta_t=zeta_t, zeta_sp=zeta_sp, zeta_a=zax, n=n, V=V, V_pi=V_pi, loc_error=loc_error, lamb=lamb)

        lims = np.array([-1, 1]) * 10

        if self.full_tests:
            norm = quad(integrate_me, lims[0], lims[1], epsabs=tol, epsrel=tol)[0]
            true_norm = 1

            # Check value
            self.assertTrue(np.isclose(norm, true_norm, rtol=self.rel_tol, atol=self.tol),
                            "Active force posterior normalization test failed for 1D posteriors in 2D. Obtained norm = %.8g did not match the expected norm = %.8g" % (norm, true_norm))

        # >> Test that 1D zeta_ax posterior is normalized in 2D without localization error <<
        zeta_t = np.asarray([0.7, 0.4])
        zeta_sp = np.asarray([0.8, 0.6])
        n = np.asarray(20)
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        loc_error = 0
        V_pi = u * V
        lamb = 'marg'
        tol = 1e-4

        # Wrapper to get just a function of zeta_ax, zeta_ay
        def integrate_me(zax):
            return calculate_one_1D_posterior_in_2D(zeta_t=zeta_t, zeta_sp=zeta_sp, zeta_a=zax, n=n, V=V, V_pi=V_pi, loc_error=loc_error, lamb=lamb)

        lims = np.array([-1, 1]) * 10

        norm = quad(integrate_me, lims[0], lims[1], epsabs=tol, epsrel=tol)[0]
        true_norm = 1

        # Check value
        self.assertTrue(np.isclose(norm, true_norm, rtol=self.rel_tol, atol=self.tol),
                        "Active force posterior normalization test failed for 1D posteriors in 2D. Obtained norm = %.8g did not match the expected norm = %.8g" % (norm, true_norm))

        # >> Test that 1D zeta_ax PRIOR is normalized in 2D with localization error <<
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        sigma2 = 0.1**2.0
        V_pi = u * V
        rtol = 1e-6
        atol = 1e-4

        def integrate_me(zax):
            return calculate_one_1D_prior_in_2D(
                zeta_a=zax, V_pi=V_pi, sigma2=sigma2)

        norm = quad(integrate_me, -np.inf, np.inf, epsrel=rtol)[0]
        true_norm = 1
        self.assertTrue(np.isclose(norm, true_norm, rtol=rtol, atol=atol),
                        "Active force prior normalization test failed for 1D posteriors in 2D. Obtained norm = %.8g did not match the expected norm = %.8g" % (norm, true_norm))

        # >> Test that 1D zeta_ax PRIOR is normalized in 2D without localization error <<
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        sigma2 = 0
        V_pi = u * V
        rtol = 1e-6
        atol = 1e-4

        def integrate_me(zax):
            return calculate_one_1D_prior_in_2D(
                zeta_a=zax, V_pi=V_pi, sigma2=sigma2)

        norm = quad(integrate_me, -np.inf, np.inf, epsrel=rtol)[0]
        true_norm = 1
        self.assertTrue(np.isclose(norm, true_norm, rtol=rtol, atol=atol),
                        "Active force prior normalization test failed for 1D posteriors in 2D. Obtained norm = %.8g did not match the expected norm = %.8g" % (norm, true_norm))

    def test_diffusivity_posterior(self):
        from .get_D_posterior import get_D_posterior, get_MAP_D
        # >> Test  diffusivity posterior norm with localization error <<
        zeta_t = np.asarray([0.7, 0.4])
        zeta_sp = np.asarray([0.8, 0.6])
        n = np.asarray(20)
        V = np.asarray(0.8 ** 2.0)
        u = np.asarray(0.95)
        loc_errors = [0, 0.1**2.0]
        V_pi = u * V
        tol = 1e-16
        dt = 0.04
        lims = np.array([0, 1e3])

        for dim in range(1, 3):
            for loc_error in loc_errors:

                # Use MAP D estimate as an integration breakpoint
                MAP_D = get_MAP_D(n=n, zeta_t=zeta_t, V=V, V_pi=V_pi,
                                  dt=dt, sigma2=loc_error, dim=2)
                posterior = get_D_posterior(n=n, zeta_t=zeta_t, V=V,
                                            V_pi=V_pi, dt=dt, sigma2=loc_error, dim=2)

                norm = quad(posterior, lims[0], lims[1], points=[MAP_D], epsabs=tol, epsrel=tol)[0]
                true_norm = 1
                self.assertTrue(np.isclose(norm, true_norm, rtol=self.rel_tol, atol=self.tol),
                                "{dim}D diffusivity posterior normalization test with localization error = {loc_error} failed in 2D. Obtained norm = {norm:.8g} did not match the expected norm = {true_norm:.8g}".format(dim=dim, loc_error=loc_error, norm=norm, true_norm=true_norm))


# # A dirty fix for a weird bug in unittest
# if __name__ == '__main__':
#     freeze_support()
# unittest.main()



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from .abc import *
from tramway.core import *
import itertools
import numpy as np
import pandas as pd


class BaseTracker(AnalyzerNode):
    __slots__ = ()
    @property
    def spt_data(self):
        if not self._eldest_parent.spt_data.initialized:
            self._eldest_parent.spt_data.from_tracker()
        return self._eldest_parent.spt_data
    @property
    def frame_interval(self):
        return self.spt_data.frame_interval
    @frame_interval.setter
    def frame_interval(self, dt):
        self.spt_data.frame_interval = dt
    @property
    def dt(self):
        return self.spt_data.dt
    @dt.setter
    def dt(self, dt):
        self.spt_data.dt = dt
    @property
    def localization_error(self):
        return self.spt_data.localization_error
    @localization_error.setter
    def localization_error(self, err):
        self.spt_data.localization_error = err
    @property
    def localization_precision(self):
        return self.spt_data.localization_precision
    @localization_precision.setter
    def localization_precision(self, sigma):
        self.spt_data.localization_precision = sigma
    @property
    def temperature(self):
        return self.spt_data.temperature
    @temperature.setter
    def temperature(self, T):
        self.spt_data.temperature = T


class SingleParticleTracker(BaseTracker):
    __slots__ = ()
    def track(self, locations, register=False, source=None):
        if isinstance(locations, str):
            loc_file = locations
            locations = load_xyt(loc_file, columns=list('xyt'))
        else:
            loc_file = None

        trajectory_index = pd.DataFrame(np.full((len(locations),1), 1, dtype=int), columns=['n'])
        trajectories = trajectory_index.join(locations)

        if register:
            self.spt_data.add_tracked_data(trajectories, loc_file if source is None else source)

        return trajectories

Tracker.register(SingleParticleTracker)


def _tolist(a):
    a = a.tolist()
    if not isinstance(a, list):
        a = [ a ]
    return a

class NonTrackingTracker(BaseTracker):
    """ Non-tracking tracker.
    """
    __slots__ = ('_high_diffusivity','_large_length')
    def __init__(self, **kwargs):
        BaseTracker.__init__(self, **kwargs)
        self._high_diffusivity = None
        self._large_length = None
    @property
    def sigma(self):
        return self.localization_precision
    @property
    def estimated_high_diffusivity(self):
        return self._high_diffusivity
    @estimated_high_diffusivity.setter
    def estimated_high_diffusivity(self, D):
        self._high_diffusivity = D
    @property
    def estimated_large_length(self):
        if self._large_length is None and \
                self.estimated_high_diffusivity is not None and \
                self.dt is not None:
            self._large_length = 2.*np.sqrt(2.*self.estimated_high_diffusivity*self.dt)
        return self._large_length
    @estimated_large_length.setter
    def estimated_large_length(self, length):
        self._large_length = length
    ###
    def track(self, locations, register=False, source=None):
        import tramway.tracking.track_non_track.file_processing_loc as nt
        images = self._eldest_parent.images
        extra_cols = None
        if isinstance(locations, str):
            loc_file = locations
            locations = load_xyt(loc_file, columns=list('xyt'))
        else:
            loc_file = None
            loc = locations
            locations = locations[list('xyt')]
            extra_cols = [ col for col in loc if col not in locations ]

        movie_per_frame, n_unique = nt.convert_to_list(locations.values)
        dt_theo = self.dt
        t_init = self.dt * 1
        if images.initialized:
            t_end = self.dt * images.n_frames
        else:
            t_end = self.dt * n_unique

        sigma = self.localization_precision
        D_high = self.estimated_high_diffusivity
        length_high = self.estimated_large_length

        if sigma is None:
            raise AttributeError('attribute localization_precision is not set')
        if D_high is None:
            raise AttributeError('attribute estimated_high_diffusivity is not set')

        trajectories = {}

        current_trajectory_index = np.uint32(1)
        currently_assigned = set()

        for frame_index in range(n_unique-1):

            C = nt.get_cost_function(frame_index, movie_per_frame)

            try:
                C_eff,_,_,_,row_eff,col_eff,M,N,n_row_eff,n_col_eff,anomaly = \
                    nt.correct_cost_function(C, length_high)
            except IndexError:
                #print(C, length_high)
                currently_assigned = set()
                continue

            _, row, col = \
                    nt.get_assigment_matrix_from_reduced_cost(C_eff, row_eff, col_eff,
                        M, N, n_col_eff, n_row_eff, anomaly)

            assert isinstance(row, np.ndarray)
            if row.size == 0:
                currently_assigned = set()
                continue

            if currently_assigned:
                source_ = movie_per_frame[frame_index]
                destination = movie_per_frame[frame_index+1]
                row_to_col = np.full(len(source_), -1, dtype=int)
                row_to_col[row] = col
                if row.size == 1:
                    row, col = _tolist(row), _tolist(col)
                newly_assigned = set(row)
                new_assignment = np.zeros(len(destination), dtype=np.uint32)
                growing_trajectory = currently_assigned & newly_assigned
                for i in growing_trajectory:
                    traj_index = current_assignment[i]
                    assert 0<traj_index
                    j = row_to_col[i]
                    trajectories[traj_index].append(destination[[j]])
                    new_assignment[j] = traj_index
                for i in newly_assigned - growing_trajectory:
                    traj_index = current_trajectory_index
                    current_trajectory_index += 1
                    j = row_to_col[i]
                    trajectories[traj_index] = [ source_[[i]], destination[[j]] ]
                    new_assignment[j] = traj_index
                current_assignment = new_assignment

            else:
                if row.size == 1:
                    row, col = _tolist(row), _tolist(col)
                source_ = movie_per_frame[frame_index][row]
                destination = movie_per_frame[frame_index+1][col]
                n = current_trajectory_index
                new_trajectory_indices = np.arange(n, n+len(col), dtype=np.uint32)
                current_trajectory_index += len(col)
                current_assignment = np.zeros(len(movie_per_frame[frame_index+1]), dtype=np.uint32)
                current_assignment[col] = new_trajectory_indices
                for k in np.arange(len(col), dtype=np.uint32):
                    trajectories[new_trajectory_indices[k]] = [ source_[[k]], destination[[k]] ]

            try:
                currently_assigned = set(col)
            except TypeError:
                print(col)

        trajectory_indices, trajectory_coordinates = [], []
        for n in range(1, current_trajectory_index):
            traj = trajectories[np.uint32(n)]
            trajectory_indices.append(np.full((sum([len(t) for t in traj]),1), n))
            trajectory_coordinates.append(traj)
        trajectory_indices = np.vstack(trajectory_indices)
        trajectory_coordinates = np.vstack(list(itertools.chain(*trajectory_coordinates)))
        trajectories = pd.DataFrame(trajectory_indices, columns=['n']).join(
                pd.DataFrame(trajectory_coordinates, columns=list('xyt')))

        if extra_cols:
            trajectories = join_by_values(trajectories, loc, list('xyt'))

        if register:
            self.spt_data.add_tracked_data(trajectories, filepath=loc_file if source is None else source)

        return trajectories

Tracker.register(NonTrackingTracker)


def join_by_values(sample, full, join_cols):
    assert not any([ col in full for col in sample if col not in join_cols ])
    ret = sample.merge(full, on=join_cols, how='left', validate='1:1')
    assert len(ret) == len(sample)
    return ret


class TrackerInitializer(Initializer):
    """ Initializer class for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.tracker` main attribute.

    The :attr:`~tramway.analyzer.RWAnalyzer.tracker` attribute self-modifies
    on calling any of the *from_...* methods.
    """
    __slots__ = ()
    def from_single_particle(self):
        """ Considers every single molecule localization datablocks
        as single trajectories. 

        See also :class:`SingleParticleTracker`."""
        self.specialize( SingleParticleTracker )
    def from_non_tracking(self):
        """ *Non-tracking* tracker.
        
        See also :class:`NonTrackingTracker`."""
        self.specialize( NonTrackingTracker )


__all__ = [ 'Tracker', 'TrackerInitializer', 'SingleParticleTracker', 'NonTrackingTracker' ]




# -*- coding: utf-8 -*-

# Copyright © 2018-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import *
from tramway.core.hdf5 import *
from tramway.core.exceptions import SideEffectWarning
import tramway.core.analyses.abc as abc
from tramway.inference import *
from ..base import *
import tramway.inference as inference # inference.plugins
import tramway.tessellation.time
import tramway.inference.time
from tramway.helper.tessellation import *
from warnings import warn
import os.path
import time
import collections
import traceback
from .old import *
_clip
# no module-wide matplotlib import for head-less usage of `infer`
# in the case matplotlib's backend is interactive


class Infer(Helper):
    def __init__(self):
        Helper.__init__(self)
        self.plugins = inference.plugins
        self.cells = None
        self.input_maps = None

    def prepare_data(self, input_data, types=None, labels=None, metadata=True, verbose=None, \
            output_file=None, **kwargs):
        Cells = (Partition, Distributed)
        if types is None:
            if labels is None and self.inplace:
                types = (Cells, (Maps, ))
            else:
                types = (Cells, )
        data = Helper.prepare_data(self, input_data, labels, types, metadata, verbose, **kwargs)
        if types[1:]:
            self.cells, self.input_maps = data
            data = self.input_maps
        elif isinstance(input_data, Cells):
            self.cells = data
        else:
            assert isinstance(data, tuple)
            self.cells, = data
            if self.explicit_input_label and self.label_is_absolute(self.input_label):
                analysis = self.analyses
                for label in self.input_label:
                    analysis = analysis[label]
                if lazytype(analysis._data) is Maps:
                    self.input_maps = data = analysis.data
        if self.input_file and output_file and (labels or labels == 0):
            self.analyses = extract_analyses(self.analyses, labels)
        return data

    def distribute(self, new_cell=None, new_group=None, cell_sampling=None,
            include_empty_cells=False, merge_threshold_count=False,
            max_cell_count=None, dilation=None, grad=None, rgrad=None):
        cells = self.cells
        if isinstance(cells, Distributed):
            _map = cells
        else:
            if not isinstance(cells, Partition) or cells.tessellation is None:
                raise ValueError('no cells found')
            has_time_linking = isinstance(cells.tessellation, tramway.tessellation.time.TimeLattice) \
                    and bool(cells.tessellation.time_dimension)
            # prepare the data for the inference
            distributed_kwargs = {}
            if new_cell is None and has_time_linking:
                new_cell = tramway.inference.time.DynamicTranslocations
            if new_group is None:
                if has_time_linking:
                    new_group = tramway.inference.time.DynamicCells
                elif merge_threshold_count:
                    new_group = DistributeMerge
                    distributed_kwargs['new_group_kwargs'] = \
                        {'min_location_count': merge_threshold_count}
                else:
                    new_group = Distributed
            if grad is not None or rgrad is not None:
                if not callable(grad):
                    if grad == 'grad1':
                        grad = grad1
                    elif grad == 'gradn':
                        grad = gradn
                    elif grad is not None:
                        raise ValueError('unsupported gradient')
                        grad = None
                if not callable(rgrad):
                    if rgrad == 'delta0':
                        rgrad = delta0
                    elif rgrad == 'delta1':
                        rgrad = delta1
                    elif rgrad == 'delta0_without_scaling':
                        rgrad = delta0_without_scaling
                    elif rgrad is not None:
                        raise ValueError("unsupported regularizing 'gradient'")
                        rgrad = None
                if grad is None:
                    class Distr(new_group):
                        def local_variation(self, *args, **kwargs):
                            return rgrad(self, *args, **kwargs)
                elif rgrad is None:
                    class Distr(new_group):
                        def grad(self, *args, **kwargs):
                            return grad(self, *args, **kwargs)
                else:
                    class Distr(new_group):
                        def grad(self, *args, **kwargs):
                            return grad(self, *args, **kwargs)
                        def local_variation(self, *args, **kwargs):
                            return rgrad(self, *args, **kwargs)
                new_group = Distr
            detailled_map = distributed(cells, new_cell=new_cell, new_group=new_group,
                    include_empty_cells=include_empty_cells, **distributed_kwargs)

            if cell_sampling is None:
                try:
                    cell_sampling = self.setup['cell_sampling']
                except KeyError:
                    pass
            multiscale = cell_sampling in ['individual', 'group', 'connected']
            if multiscale and max_cell_count is None:
                if cell_sampling == 'individual':
                    max_cell_count = 1
                #else: # adaptive scaling is no longer default
                #       max_cell_count = 20
            if cell_sampling == 'connected':
                multiscale_map = detailled_map.group(connected=True)
                _map = multiscale_map
            elif max_cell_count:
                if dilation is None:
                    if cell_sampling == 'individual':
                        dilation = 0
                    else:
                        dilation = 2
                multiscale_map = detailled_map.group(max_cell_count=max_cell_count, \
                    adjacency_margin=dilation)
                _map = multiscale_map
            else:
                _map = detailled_map
        return _map

    def overload_cells(self, cells):
        if self.input_maps is None:
            input_features = ()
        else:
            input_features = []
            for input_feature in self.input_maps.features:
                # check `input_feature` can be an identifier
                try:
                    class _BreakMe(object):
                        __slots__ = input_feature
                except TypeError:
                    pass
                else:
                    input_features.append(input_feature)
            input_features = tuple(input_features)
            maps = { v: self.input_maps[v] for v in input_features }
        output_features = self.setup.get('returns', [])
        if isinstance(output_features, (tuple, list, frozenset, set)):
            output_features = tuple(output_features)
        else:
            output_features = (output_features,)
        if input_features or output_features:
            any_cell = cells.any_cell()
            cell_type = type(any_cell)
            try:
                attrs = any_cell.__dict__
            except AttributeError:
                attrs = Lazy.__slots__ + Local.__slots__ + Cell.__slots__ + cell_type.__slots__
                conflicting_names = set(output_features) & set(attrs)
                if conflicting_names:
                    warn('output feature name is also an existing attribute: {}'.format(list(conflicting_names)), SideEffectWarning)
            class OverloadedCell(cell_type):
                __slots__ = input_features + output_features
                def __init__(self, cell, **kwargs):
                    for attr in attrs:
                        setattr(self, attr, getattr(cell, attr))
                    for attr in input_features:
                        setattr(self, attr, kwargs[attr])
                    for attr in output_features:
                        setattr(self, attr, None)
            kwargs = {}
            overloaded_cells = {}
            for i in cells:
                cell = cells[i]
                if self.input_maps is not None:
                    for k in maps:
                        try:
                            val = maps[k].loc[i].values
                        except KeyError:
                            val = None
                        else:
                            if np.isscalar(val):
                                val = val.tolist()
                        kwargs[k] = val
                overloaded_cells[i] = OverloadedCell(cell, **kwargs)
            cells.cells = overloaded_cells
        return cells

    def insert_mappable_cells(self, cells, label=None, comment=None, anchor=None):
        if anchor is None:
            anchor = self.cells
        # backup output_label
        output_label = self.output_label
        if self.label_is_absolute(output_label):
            raise NotImplementedError('absolute output paths are not supported in combination with store_distributed')
        self.output_label = None
        # insert the mappable cells
        self.input_label = self.insert_analysis(cells, anchor=anchor, label=label, comment=comment)
        # restore output_label
        self.output_label = output_label

    def infer(self, cells, worker_count=None, profile=None, min_diffusivity=None, \
            localization_error=None, sigma=None, sigma2=None, \
            diffusivity_prior=None, potential_prior=None, jeffreys_prior=None, rgrad=None, \
            comment=None, verbose=None, snr_extensions=False, **kwargs):
        if verbose is None:
            verbose = self.verbose
        mode = self.name
        runtime = time.time()

        args = self.setup.get('arguments', {})
        for arg in ('localization_error', 'sigma', 'sigma2', 'diffusivity_prior', 'potential_prior',
                'jeffreys_prior', 'min_diffusivity', 'worker_count', 'verbose'):
            try:
                args[arg]
            except KeyError:
                pass
            else:
                val = eval(arg)
                if val is not None:
                    kwargs[arg] = val
        diffusion_prior = kwargs.pop('diffusion_prior', None)
        if diffusion_prior is not None:
            kwargs['diffusivity_prior'] = diffusion_prior

        if 'returns' in self.setup:
            kwargs['returns'] = self.setup['returns']
        if profile:
            kwargs['profile'] = profile
        if rgrad:
            kwargs['rgrad'] = rgrad

        try:
            _fun = getattr(self.module, self.setup['infer'])
        except KeyError:
            _fun = self._infer

        if self.input_is_partition:

            if cells is None:
                cells = self.cells
            x = _fun(cells, **kwargs)

        else:

            x = cells.run(_fun, **kwargs)

        ret = {}
        if isinstance(x, tuple):
            if isinstance(x[1], pd.DataFrame):
                maps = Maps(x[0], mode=mode, posteriors=x[1])
                if x[2:]:
                    try:
                        maps.other = x[2:] # Python 3 only
                    except:
                        warn('failed to store output arguments: {}'.format(x[2:]), RuntimeWarning)
                        if verbose:
                            print(traceback.format_exc())
            else:
                maps = Maps(x[0], mode=mode)
                ret = x[1]
                if x[2:]:
                    warn('ignoring output arguments: {}'.format(x[2:]), RuntimeWarning)
        else:
            maps = Maps(x, mode=mode)

        for p in kwargs:
            if p not in ['worker_count', 'profile', 'returns']:
                setattr(maps, p, kwargs[p])

        runtime = time.time() - runtime
        if verbose:
            print('{} mode: elapsed time: {}ms'.format(mode, int(round(runtime*1e3))))
        maps.runtime = runtime

        for attr in ret:
            maps.defattr(attr, ret[attr])

        if snr_extensions:
            if self.input_is_partition:
                raise TypeError
            maps = inference.snr.add_snr_extensions(cells, maps, get_grad_kwargs(**kwargs))

        if self.analyses is not None:
            self.insert_analysis(maps, comment=comment)

        return maps

    def plugin(self, name, plugins=None, verbose=None, func=None, **kwargs):
        if func is not None:
            if plugins is None:
                plugins = {}
            elif name in plugins:
                warn('plugin `{}` will be overwritten'.format(name), RuntimeWarning)
            setup = kwargs
            setup['infer'] = 'func'
            Plugin = collections.namedtuple('Plugin', ('func',))
            module = Plugin(func)
            plugins[name] = (setup, module)
        Helper.plugin(self, name, plugins, verbose)

    @property
    def input_is_partition(self):
        input_type = self.setup.get('input_type', None)
        return input_type and input_type.lower().endswith('partition')


def infer1(cells, mode='degraded.d', output_file=None, partition={}, verbose=False, \
    localization_error=None, diffusivity_prior=None, potential_prior=None, jeffreys_prior=None, \
    max_cell_count=None, dilation=None, worker_count=None, min_diffusivity=None, \
    store_distributed=False, new_cell=None, new_group=None, constructor=None, \
    include_empty_cells=False, cell_sampling=None, merge_threshold_count=False, \
    grad=None, rgrad=None, input_label=None, output_label=None, comment=None, \
    return_cells=None, profile=None, overwrite=None, inplace=False, \
    priorD=None, priorV=None, force=None, **kwargs):
    """
    Inference helper.

    Arguments:

        cells (str or Partition or tramway.core.analyses.base.Analyses):
            data partition or path to partition file

        mode (str or callable): plugin name; see for example
            :mod:`~tramway.inference.d` (``'d'``),
            :mod:`~tramway.inference.df` (``'df'``),
            :mod:`~tramway.inference.dd` (``'dd'``),
            :mod:`~tramway.inference.dv` (``'dv'``);
            can be also a function suitable for :meth:`~tramway.helper.inference.base.Distributed.run`

        output_file (str): desired path for the output map file

        verbose (bool or int): verbosity level

        localization_error/sigma (float): localization error (see also sigma2)

        diffusivity_prior/diffusion_prior (float): hyperparameter of the prior on the
            diffusivity/diffusion

        potential_prior (float): hyperparameter of the prior on the potential energy

        jeffreys_prior (float): Jeffreys' prior

        max_cell_count (int): if defined, divide the mesh into convex subsets of cells

        dilation (int): overlap of side cells if `max_cell_count` is defined

        worker_count (int): number of parallel processes to be spawned

        min_diffusivity (float): (possibly negative) lower bound on local diffusivities

        store_distributed (bool): store the :class:`~tramway.inference.base.Distributed` object
            in the map file

        new_cell (callable): see also :func:`~tramway.inference.base.distributed`

        new_group (callable): see also :func:`~tramway.inference.base.distributed`

        constructor (callable): **deprecated**; see also :func:`~tramway.inference.base.distributed`;
            please use `new_group` instead

        include_empty_cells (bool): do not discard cells with no (trans-)locations

        cell_sampling (str): either ``None``, ``'individual'``, ``'group'`` or
            ``'connected'``; may ignore `max_cell_count` and `dilation`

        merge_threshold_count (int):
            Merge cells that are have a number of (trans-)locations lower than the
            number specified; each smaller cell is merged together with the nearest
            large-enough neighbour cell.

        grad (callable or str): spatial gradient function; admits a callable (see
            :meth:`~tramway.inference.base.Distributed.grad`) or any of '*grad1*',
            '*gradn*'

        rgrad (None or 'delta' or 'delta1'): alternative space/time "gradient" implementation
            for the regularization priors;
            if defined, inference modes that would otherwise use
            :meth:`~tramway.inference.base.Distributed.grad` will use
            :meth:`~tramway.inference.base.Distributed.local_variation` instead

        input_label (list): label path to the input :class:`~tramway.tessellation.base.Tessellation`
            object in `cells` if the latter is an :class:`~tramway.core.analyses.base.Analyses`
            or filepath

        output_label (str): label for the resulting analysis instance

        comment (str): description message for the resulting analysis

        return_cells (bool): return a tuple with a :class:`~tramway.tessellation.base.Partition`
            object as extra element

        profile (bool or str): profile each child job if any;
            if `str`, dump the output stats into *.prof* files;
            if `tuple`, print a report with :func:`~pstats.Stats.print_stats` and
            tuple elements as input arguments.

        overwrite (bool): if an implicit output file already exists, overwrite it.

        inplace (bool): replace the input analysis by the output one.

        snr_extensions (bool): add snr extensions for Bayes factor calculation.

    Returns:

        Maps or pandas.DataFrame or tuple:

    `priorD` and `priorV` are legacy arguments.
    They are deprecated and `diffusivity_prior`/`diffusion_prior` and `potential_prior` respectively
    should be used instead.
    """
    if bool(partition):
        warn('the `partition` argument is ignored and will be removed', DeprecationWarning)

    helper = Infer()
    helper.verbose = verbose
    helper.labels(input_label=input_label, output_label=output_label, inplace=inplace, comment=comment)
    cells = helper.prepare_data(cells, labels=input_label, metadata=not kwargs.pop('disable_metadata',None))

    if mode in ('D', 'DF', 'DD', 'DV'):
        mode = mode.lower()
        #warn('inference mode: please use degraded.{} instead'.format(mode), PendingDeprecationWarning)
    helper.plugin(mode)

    if helper.input_is_partition:
        _map = None
    else:

        if constructor is not None:
            warn('the `constructor` argument is deprecated; please use `new_group` instead',
                    DeprecationWarning)
        _map = helper.distribute(new_cell=new_cell, \
                new_group=constructor if new_group is None else new_group, cell_sampling=cell_sampling, \
                include_empty_cells=include_empty_cells, \
                merge_threshold_count=merge_threshold_count, max_cell_count=max_cell_count, \
                dilation=dilation, grad=grad, rgrad=rgrad)
        _map = helper.overload_cells(_map)

        if store_distributed:
            helper.insert_mappable_cells(_map, anchor=cells, \
                label=None if isinstance(store_distributed, bool) else store_distributed)

    if diffusivity_prior is None and priorD is not None:
        diffusivity_prior = priorD
    if potential_prior is None and priorV is not None:
        potential_prior = priorV
    maps = helper.infer(_map, worker_count=worker_count, profile=profile, \
        min_diffusivity=min_diffusivity, localization_error=localization_error, \
        diffusivity_prior=diffusivity_prior, potential_prior=potential_prior, \
        jeffreys_prior=jeffreys_prior, rgrad=rgrad, **kwargs)

    if overwrite is None and force is not None:
        warn('`force` is deprecated; please use `overwrite` instead', PendingDeprecationWarning)
        overwrite = force
    helper.save_analyses(output_file, force=overwrite)

    if return_cells == True: # NOT `is`
        return maps, cells
    elif return_cells == False:
        return maps.maps # old
    elif helper.input_file and return_cells == 'first':
        return cells, mode, maps.maps # old
    else:
        return maps # new


def infer0(cells, mode='D', output_file=None, partition={}, verbose=False, \
    localization_error=None, diffusivity_prior=None, potential_prior=None, jeffreys_prior=None, \
    max_cell_count=None, dilation=None, worker_count=None, min_diffusivity=None, \
    store_distributed=False, new_cell=None, new_group=None, constructor=None, cell_sampling=None, \
    merge_threshold_count=False, \
    grad=None, priorD=None, priorV=None, input_label=None, output_label=None, comment=None, \
    return_cells=None, profile=None, force=False, **kwargs):
    """
    Inference helper.

    Arguments:

        cells (str or Partition or tramway.core.analyses.base.Analyses):
            data partition or path to partition file

        mode (str or callable): plugin name; see for example
            :mod:`~tramway.inference.d` (``'d'``),
            :mod:`~tramway.inference.df` (``'df'``),
            :mod:`~tramway.inference.dd` (``'dd'``),
            :mod:`~tramway.inference.dv` (``'dv'``);
            can be also a function suitable for :meth:`~tramway.helper.inference.base.Distributed.run`

        output_file (str): desired path for the output map file

        partition (dict): keyword arguments for :func:`~tramway.helper.tessellation.find_partition`
            if `cells` is a path; **deprecated**

        verbose (bool or int): verbosity level

        localization_error (float): localization error

        diffusivity_prior (float): prior diffusivity

        potential_prior (float): prior potential

        jeffreys_prior (float): Jeffreys' prior

        max_cell_count (int): if defined, divide the mesh into convex subsets of cells

        dilation (int): overlap of side cells if `max_cell_count` is defined

        worker_count (int): number of parallel processes to be spawned

        min_diffusivity (float): (possibly negative) lower bound on local diffusivities

        store_distributed (bool): store the :class:`~tramway.inference.base.Distributed` object
            in the map file

        new_cell (callable): see also :func:`~tramway.inference.base.distributed`

        new_group (callable): see also :func:`~tramway.inference.base.distributed`

        constructor (callable): **deprecated**; see also :func:`~tramway.inference.base.distributed`;
            please use `new_group` instead

        cell_sampling (str): either ``None``, ``'individual'``, ``'group'`` or
            ``'connected'``; may ignore `max_cell_count` and `dilation`

        merge_threshold_count (int):
            Merge cells that are have a number of (trans-)locations lower than the
            number specified; each smaller cell is merged together with the nearest
            large-enough neighbour cell.

        grad (callable or str): spatial gradient function; admits a callable (see
            :meth:`~tramway.inference.base.Distributed.grad`) or any of '*grad1*',
            '*gradn*'

        input_label (list): label path to the input :class:`~tramway.tessellation.base.Tessellation`
            object in `cells` if the latter is an :class:`~tramway.core.analyses.base.Analyses`
            or filepath

        output_label (str): label for the resulting analysis instance

        comment (str): description message for the resulting analysis

        return_cells (bool): return a tuple with a :class:`~tramway.tessellation.base.Partition`
            object as extra element

        profile (bool or str): profile each child job if any;
            if `str`, dump the output stats into *.prof* files;
            if `tuple`, print a report with :func:`~pstats.Stats.print_stats` and
            tuple elements as input arguments.

    Returns:

        Maps or pandas.DataFrame or tuple:

    `priorD` and `priorV` are legacy arguments.
    They are deprecated and `diffusivity_prior` and `potential_prior` should be used instead
    respectively.
    """
    if verbose:
        inference.plugins.verbose = True

    if diffusivity_prior is None and priorD is not None:
        diffusivity_prior = priorD
    if potential_prior is None and priorV is not None:
        potential_prior = priorV

    input_file = None
    all_analyses = analysis = None
    if isinstance(cells, str):
        try:
            input_file = cells
            all_analyses = load_rwa(input_file, lazy=True)
            if output_file and output_file == input_file:
                all_analyses = extract_analysis(all_analyses, input_label)
            cells = None
        except KeyError:
            raise
            # legacy format
            input_file, cells = find_partition(cells, **partition)
            if cells is None:
                raise ValueError('no cells found')
        if verbose:
            print('loading file: {}'.format(input_file))
    elif isinstance(cells, abc.Analyses):
        all_analyses, cells = cells, None
    elif not isinstance(cells, Partition):
        raise TypeError('wrong type for argument `cells`')

    if cells is None:
        if not all_analyses:
            raise ValueError('no cells found')
        if not input_label:
            labels = tuple(all_analyses.labels)
            if labels[1:]:
                raise ValueError('multiple instances; input_label is required')
            input_label = labels[-1]
        if isinstance(input_label, (tuple, list)):
            if input_label[1:]:
                analysis = all_analyses
                for label in input_label:#[:-1]
                    analysis = analysis[label]
                cells = analysis.data
                #analysis = analysis[input_label[-1]]
                #if not isinstance(cells, Partition):
                #       cells = analysis.data
            else:
                input_label = input_label[0]
        if cells is None:
            analysis = all_analyses[input_label]
            cells = analysis.data
        if not isinstance(cells, (Partition, Distributed)):
            raise ValueError('cannot find cells at the specified label')
    elif all_analyses is None:
        all_analyses = Analyses(cells.points)
        assert analysis is None
        analysis = Analyses(cells)
        all_analyses.add(analysis)
        assert input_label is None
        input_label = tuple(all_analyses.labels)

    if mode in ('D', 'DF', 'DD', 'DV'):
        mode = mode.lower()
    setup, module = inference.plugins[mode]

    if isinstance(analysis.data, Distributed):
        _map = analysis.data
    else:

        if cells is None or cells.tessellation is None:
            raise ValueError('no cells found')

        # prepare the data for the inference
        distributed_kwargs = {}
        if new_group is None:
            if constructor is None:
                if merge_threshold_count:
                    new_group = DistributeMerge
                    distributed_kwargs['new_group_kwargs'] = \
                        {'min_location_count': merge_threshold_count}
                else:
                    new_group = Distributed
            else:
                new_group = constructor
        if grad is not None:
            if not callable(grad):
                if grad == 'grad1':
                    grad = grad1
                elif grad == 'gradn':
                    grad = gradn
                else:
                    raise ValueError('unsupported gradient')
                    grad = None
            if grad is not None:
                class Distr(new_group):
                    def grad(self, *args, **kwargs):
                        return grad(self, *args, **kwargs)
                new_group = Distr
        detailled_map = distributed(cells, new_cell=new_cell, new_group=new_group,
                **distributed_kwargs)

        if cell_sampling is None:
            try:
                cell_sampling = setup['cell_sampling']
            except KeyError:
                pass
        multiscale = cell_sampling in ['individual', 'group', 'connected']
        if multiscale and max_cell_count is None:
            if cell_sampling == 'individual':
                max_cell_count = 1
            #else: # adaptive scaling is no longer default
            #       max_cell_count = 20
        if cell_sampling == 'connected':
            multiscale_map = detailled_map.group(connected=True)
            _map = multiscale_map
        elif max_cell_count:
            if dilation is None:
                if cell_sampling == 'individual':
                    dilation = 0
                else:
                    dilation = 2
            multiscale_map = detailled_map.group(max_cell_count=max_cell_count, \
                adjacency_margin=dilation)
            _map = multiscale_map
        else:
            _map = detailled_map

        if store_distributed:
            if output_label is None:
                output_label = analysis.autoindex()
            analysis.add(Analysis(_map), label=output_label)
            analysis = analysis[output_label]
            output_label = None

    runtime = time.time()

    if mode in inference.plugins:

        args = setup.get('arguments', {})
        for arg in ('localization_error', 'diffusivity_prior', 'potential_prior',
                'jeffreys_prior', 'min_diffusivity', 'worker_count', 'verbose'):
            try:
                args[arg]
            except KeyError:
                pass
            else:
                val = eval(arg)
                if val is not None:
                    kwargs[arg] = val
        if profile:
            kwargs['profile'] = profile
        x = _map.run(getattr(module, setup['infer']), **kwargs)

    else:
        raise ValueError("unknown '{}' mode".format(mode))

    if isinstance(x, tuple):
        maps = Maps(x[0], mode=mode, posteriors=x[1])
        if x[2:]:
            maps.other = x[2:] # Python 3 only
    else:
        maps = Maps(x, mode=mode)

    for p in kwargs:
        if p not in ['worker_count']:
            setattr(maps, p, kwargs[p])
    analysis.add(Analyses(maps), label=output_label, comment=comment)

    runtime = time.time() - runtime
    if verbose:
        print('{} mode: elapsed time: {}ms'.format(mode, int(round(runtime*1e3))))
    maps.runtime = runtime

    if input_file and not output_file:
        output_file = input_file

    if output_file:
        # store the result
        save_rwa(output_file, all_analyses, verbose, force=input_file == output_file or force)

    if return_cells == True: # NOT `is`
        return (maps, cells)
    elif return_cells == False:
        return maps
    elif input_file:
        if return_cells is not None:
            warn("3-element return value will no longer be the default; pass return_cells='first' to maintain this behavior", FutureWarning)
        return (cells, mode, x)
    else:
        return x


def map_plot1(maps, cells=None, clip=None, output_file=None, fig_format=None, \
    figsize=None, dpi=None, aspect=None, show=None, verbose=False, \
    alpha=None, point_style=None, feature=None, variable=None, segment=None, \
    label=None, input_label=None, mode=None, title=True, inferencemap=False, \
    use_bokeh=None, **kwargs):
    """
    Plot scalar/vector 2D maps.

    For consistency with *InferenceMAP*, the effective potential is showed
    in :math:`µm^{-1}` while -- using ``unit='std'`` -- the colorbar
    mentions :math:`k_{B}T`, whereas no conversion factor is applied.

    Arguments:

        maps (str or tramway.core.analyses.base.Analyses or pandas.DataFrame or Maps):
            maps as a path to a rwa map file,
            an analysis tree, a dataframe or a :class:`Maps`;
            filepaths and analysis trees may require `label` (or equivalently `input_label`)
            to be defined; dataframes and encapsulated maps require `cells` to be defined

        cells (Partition or Tessellation or Distributed): mesh with optional partition

        clip (float): clips map values by absolute values;
            if ``clip < 1``, it is the quantile at which to clip absolute values of the map;
            otherwise it defines: ``threshold = median + clip * (third_quartile - first_quartile)``

        output_file (str): path to output file

        fig_format (str): for example '*.png*'

        figsize (bool or (float, float)): figure size (width, height) in inches;
            `figsize` is defined if multiple figures are drawn or the figures are printed
            to files, which opens a new figure for each plot;
            this can be prevented setting `figsize` to ``False``

        dpi (int): dots per inch

        aspect (float or str): aspect ratio or '*equal*'

        show (bool or str): call :func:`~matplotlib.pyplot.show`; if ``show='draw'``, call
            :func:`~matplotlib.pyplot.draw` instead.
            `show` is ``True`` if the figures are not printed to files; to maintain this
            default behaviour in future releases, set `show` to ``True`` from now on.

        verbose (bool): verbosity level

        alpha (float): alpha value for scalar maps; useful in combination with `point_style`;
            if ``False``, the alpha value is not explicitly set

        point_style (dict): if defined, points are overlaid

        feature/variable (str): feature name (e.g. 'diffusivity', 'force')

        segment (int): segment index;
            if multiple time segments were defined, show only this segment

        label/input_label (int or str): analysis instance label

        mode (bool or str): inference mode; can be ``False`` so that mode information from
            files, analysis trees and encapsulated maps are not displayed

        title (bool or str): add titles to the figures, based on the feature name;
            from version *0.4*, the inference mode is no longer appended

        inferencemap (bool): scales the arrows wrt to cell size only; for field maps only

        xlim (array-like): min and max values for the x-axis; this argument is keyworded only

        ylim (array-like): min and max values for the y-axis; this argument is keyworded only

        zlim (array-like): min and max values for the z-axis; this argument is keyworded only;
            applies only to scalar maps that are consequently plotted in 3D

        clim (array-like): min and max values for the colormap; this argument is keyworded only

        unit (str): colorbar label; can be :const:`'std'` to take usual units,
            with space coordinates in *µm* and and times in *s*

    Extra keyword arguments may be passed to :func:`~tramway.plot.map.scalar_map_2d`,
    :func:`~tramway.plot.map.field_map_2d` and :func:`scalar_map_3d`.
    They can be dictionnaries with feature names as keys and the corresponding values for the
    parameters.

    """
    # transitional function; may diverge from map_plot0

    # get cells and maps objects from the first input argument
    input_file = None
    if isinstance(maps, tuple):
        warn('`maps` as (Partition, str, DataFrame) tuple are deprecated', DeprecationWarning)
        cells, mode, maps = maps
    elif isinstance(maps, (pd.DataFrame, Maps, pd.Series)):
        if cells is None:
            raise ValueError('`cells` is not defined')
    elif isinstance(maps, abc.Analyses):
        analyses = maps
        if label is None:
            label = input_label
        cells, maps = find_artefacts(analyses, ((Partition, Distributed), Maps), label)
    elif isinstance(maps, str): # `maps` is a file path
        input_file = maps
        if not os.path.isfile(input_file):
            raise OSError('cannot find file: {}'.format(input_file))
        if label is None:
            label = input_label
        try:
            analyses = load_rwa(input_file, lazy=True)
            #if label:
            #       analyses = extract_analysis(analyses, label)
        except KeyError:
            print(traceback.format_exc())
            from rwa import HDF5Store
            store = HDF5Store(input_file, 'r')
            store.lazy = False
            try:
                # old format
                maps = peek_maps(store, store.store)
            finally:
                store.close()
            try:
                tess_file = maps.rwa_file
            except AttributeError:
                # even older
                tess_file = maps.imt_file
            if not isinstance(tess_file, str):
                tess_file = tess_file.decode('utf-8')
            tess_file = os.path.join(os.path.dirname(input_file), tess_file)
            store = HDF5Store(tess_file, 'r')
            store.lazy = False
            try:
                cells = store.peek('cells')
                if cells.tessellation is None:
                    cells._tessellation = store.peek('_tesselation', store.store['cells'])
            finally:
                store.close()
        except ImportError:
            warn('HDF5 libraries may not be installed', ImportWarning)
        else:
            cells, maps = find_artefacts(analyses, ((Partition, Distributed), Maps), label)
    else:
        raise TypeError('unsupported type for maps: {}'.format(type(maps)))
    if isinstance(maps, Maps):
        if mode != False:
            mode = maps.mode
        maps = maps.maps
    elif isinstance(maps, pd.Series):
        maps = pd.DataFrame(maps.values, index=maps.index, columns=['unknown feature'])
    if isinstance(cells, Distributed):
        # fix for rwa-0.5 OrderedDict
        cells.cells = collections.OrderedDict((k, cells[k]) for k in range(max(cells.keys())+1) if k in cells )

    if not cells._lazy.get('bounding_box', True):
        maps = box_crop(maps, cells.bounding_box, cells.tessellation)

    xlim, ylim, zlim = kwargs.get('xlim', None), kwargs.get('ylim', None), kwargs.pop('zlim', None)
    if xlim and ylim:
        maps = box_crop(maps,
            pd.DataFrame(
                np.array([[xlim[0], ylim[0]], [xlim[1], ylim[1]]]),
                columns=['x', 'y']),
            cells.tessellation)

    unit = kwargs.pop('unit', None)
    if unit == 'std':
        # standard units are defined at multiple locations:
        # * tramway.plot.bokeh.analyzer.Controller.draw_map
        # * tramway.helper.inference.map_plot
        # * tramway.analyzer.mapper.mpl.Mpl.clabel
        unit = {'diffusivity': r'$\mu\rm{m}^2\rm{s}^{-1}$',
                'potential': r'$k_{\rm{B}}T$',
                'force': r'$k_{\rm{B}}T\mu\rm{m}^{-1}$',
                'drift': r'$\mu\rm{m}\rm{s}^{-1}$',
               }

    # identify time segments, if any
    try:
        import tramway.tessellation.time as lattice
        with_segments = isinstance(cells.tessellation, lattice.TimeLattice) \
                and cells.tessellation.spatial_mesh is not None
    except ImportError:
        with_segments = False
    if with_segments:
        if segment is None:
            raise ValueError('`segment` is required')
        elif isinstance(segment, (tuple, list)):
            if segment[1:]:
                warn('cannot plot multiple segments in a single `map_plot` call', RuntimeWarning)
            segment = segment.pop()
            print('plotting segment {}'.format(segment))
        _cells, cells = cells, cells.tessellation.split_segments(cells)[segment]
    elif segment is not None:
        warn('cannot find time segments', RuntimeWarning)

    # `mode` type may be inadequate because of loading a Py2-generated rwa file in Py3 or conversely
    if mode and not isinstance(mode, str):
        try: # Py2
            mode = mode.encode('utf-8')
        except AttributeError: # Py3
            mode = mode.decode('utf-8')

    # determine whether the figures should be printed to file or not
    print_figs = output_file or (input_file and fig_format)

    # output filenames
    if print_figs:
        if output_file:
            filename, figext = os.path.splitext(output_file)
            if fig_format:
                figext = fig_format
            elif figext and figext[1:] in fig_formats:
                figext = figext[1:]
            else:
                figext = fig_formats[0]
        else:
            figext = fig_format
            filename, _ = os.path.splitext(input_file)
        if use_bokeh is None and figext == 'html':
            use_bokeh = True

    # figure size
    new_fig = bool(figsize) or (print_figs and figsize is not False)
    if figsize in (None, True):
        if use_bokeh:
            figsize = None
        else:
            figsize = (12., 9.)

    # import graphics libraries with adequate backend
    if print_figs:
        if not use_bokeh:
            import matplotlib
            try:
                matplotlib.use('Agg') # head-less rendering (no X server required)
            except:
                pass
    if use_bokeh:
        import bokeh.plotting     as mplt
        import tramway.plot.bokeh as tplt
        if figsize:
            fig_kwargs = dict(plot_width=figsize[0], plot_height=figsize[1])
        else:
            fig_kwargs = {}
    else:
        import matplotlib.pyplot as mplt
        import tramway.plot      as tplt
        if 'figure' in kwargs:
            fig = kwargs['figure']
        if point_style is not None and 'axes' in kwargs:
            point_style['axes'] = kwargs['axes']

    # identify and plot the possibly various maps
    figs = []
    nfig = 0

    if feature is None:
        feature = variable
    all_vars = dict(splitcoord(maps.columns)) # not a defaultdict
    if isinstance(feature, (frozenset, set, tuple, list)):
        all_vars = { v: all_vars[v] for v in feature }
    elif feature is not None:
        all_vars = { feature: all_vars[feature] }

    standard_kwargs = {}
    differential_kwargs = {}
    for kw in kwargs:
        arg = kwargs[kw]
        if isinstance(arg, dict):
            keys = arg.keys()
        elif isinstance(arg, pd.Series):
            keys = arg.index
        elif isinstance(arg, pd.DataFrame):
            keys = arg.columns
        else:
            keys = ()
        if 0 < len(keys) and all(key in all_vars for key in keys):
            differential_kwargs[kw] = arg
        else:
            standard_kwargs[kw] = arg
    kwargs = standard_kwargs

    if unit:
        if isinstance(unit, dict):
            differential_kwargs['unit'] = unit
        else:
            kwargs['unit'] = unit

    scalar_vars = {'diffusivity': 'D', 'potential': 'V'}
    scalar_vars = [ (v, scalar_vars.get(v, None)) for v in all_vars if len(all_vars[v]) == 1 ]

    for col, short_name in scalar_vars:

        col_kwargs = dict(standard_kwargs)
        for kw in differential_kwargs:
            try:
                arg = differential_kwargs[kw][col]
            except KeyError:
                pass
            else:
                col_kwargs[kw] = arg

        if use_bokeh:
            if print_figs:
                mplt.output_file(output_file)
            fig = mplt.figure(**fig_kwargs)
            col_kwargs['figure'] = fig
            if point_style is not None:
                point_style['figure'] = fig
        else:
            if 'figure' in kwargs:
                pass
            elif new_fig or figs:
                fig = mplt.figure(figsize=figsize, dpi=dpi)
            else:
                fig = mplt.gcf()
        figs.append(fig)

        _map = maps[col]
        if isinstance(clip, (dict, pd.Series, pd.DataFrame)):
            try:
                __clip = clip[col]
            except:
                __clip = None
        else:
            __clip = clip
        if __clip:
            _map = _clip(_map, __clip)
        # debug
        if isinstance(maps, pd.DataFrame) and 'x' in maps.columns and col not in 'xyzt':
            _map = maps[[ col for col in 'xyzt' if col in maps.columns ]].join(_map)
        #

        # split time segments, if any
        if with_segments:
            if 'clim' not in col_kwargs:
                col_kwargs['clim'] = [_map.min(), _map.max()]
            _map = _cells.tessellation.split_frames(_map)
            try:
                _map = _map[segment]
            except IndexError:
                raise IndexError('segment index {} is out of bounds (max {})'.format(segment, len(_map)-1))

        if zlim is None:
            plot = tplt.scalar_map_2d
        else:
            plot = tplt.scalar_map_3d
            col_kwargs['zlim'] = zlim

        plot(cells, _map, aspect=aspect, alpha=alpha, **col_kwargs)

        if point_style is not None:
            points = cells.descriptors(cells.points, asarray=True) # `cells` should be a `Partition`
            if 'color' not in point_style:
                point_style['color'] = None
            tplt.plot_points(points, **point_style)

        if title and not use_bokeh:
            if isinstance(title, str):
                _title = title
            #elif mode:
            #    if short_name:
            #        _title = '{} ({} - {} mode)'.format(short_name, col, mode)
            #    else:
            #        _title = '{} ({} mode)'.format(col, mode)
            #elif short_name:
            #    _title = '{} ({})'.format(short_name, col)
            else:
                _title = '{}'.format(col)
            try:
                axes = kwargs['axes']
            except KeyError:
                mplt.title(_title)
            else:
                axes.set_title(_title)

        if print_figs and not use_bokeh:
            if maps.shape[1] == 1:
                figfile = '{}.{}'.format(filename, figext)
            elif short_name:
                figfile = '{}_{}.{}'.format(filename, short_name.lower(), figext)
            else:
                figfile = '{}_{}.{}'.format(filename, nfig, figext)
                nfig += 1
            if verbose:
                print('writing file: {}'.format(figfile))
            fig.savefig(figfile, dpi=dpi)

    vector_vars = {'force': 'F'}
    vector_vars = [ (v, vector_vars.get(v, None)) for v in all_vars if len(all_vars[v]) == 2 ]

    for name, short_name in vector_vars:
        cols = all_vars[name]

        var_kwargs = dict(standard_kwargs)
        for kw in differential_kwargs:
            try:
                arg = differential_kwargs[kw][name]
            except KeyError:
                pass
            else:
                var_kwargs[kw] = arg

        plot = tplt.field_map_2d
        if point_style is not None:
            var_kwargs['overlay'] = True
        if inferencemap:
            var_kwargs['inferencemap'] = inferencemap

        if use_bokeh:
            if print_figs:
                mplt.output_file(output_file)
            fig = mplt.figure(**fig_kwargs)
            var_kwargs['figure'] = fig
            if point_style is not None:
                point_style['figure'] = fig
        else:
            if 'figure' in kwargs:
                pass
            elif new_fig or figs:
                fig = mplt.figure(figsize=figsize, dpi=dpi)
            else:
                fig = mplt.gcf()
        figs.append(fig)

        _vector_map = maps[cols]
        if isinstance(clip, (dict, pd.Series, pd.DataFrame)):
            try:
                __clip = clip[name]
            except:
                __clip = None
        else:
            __clip = clip
        if __clip:
            _vector_map = _clip(_vector_map, __clip)

        # split time segments, if any
        if with_segments:
            if 'clim' not in var_kwargs:
                _scalar_map = _vector_map.pow(2).sum(1).apply(np.sqrt)
                var_kwargs['clim'] = [_scalar_map.values.min(), _scalar_map.values.max()]
            _vector_map = _cells.tessellation.split_frames(_vector_map)[segment]

        if point_style is not None:
            _scalar_map = _vector_map.pow(2).sum(1).apply(np.sqrt)
            tplt.scalar_map_2d(cells, _scalar_map, aspect=aspect, alpha=alpha, **var_kwargs)
            points = cells.descriptors(cells.points, asarray=True) # `cells` should be a `Partition`
            if 'color' not in point_style:
                point_style['color'] = None
            tplt.plot_points(points, **point_style)

        plot(cells, _vector_map, aspect=aspect, **var_kwargs)

        extra = None
        if short_name:
            main = short_name
            extra = name
        else:
            main = name

        if title and not use_bokeh:
            if isinstance(title, str):
                _title = title
            else:
                #if mode:
                #    if extra:
                #        extra += ' - {} mode'.format(mode)
                #    else:
                #        extra = '{} mode'.format(mode)
                #if extra:
                #    _title = '{} ({})'.format(main, extra)
                #else:
                #    _title = main
                _title = name
            try:
                axes = kwargs['axes']
            except KeyError:
                mplt.title(_title)
            else:
                axes.set_title(_title)

        if print_figs and not use_bokeh:
            if maps.shape[1] == 1:
                figfile = '{}.{}'.format(filename, figext)
            else:
                if short_name:
                    ext = short_name.lower()
                else:
                    ext = name
                figfile = '{}_{}.{}'.format(filename, ext, figext)
            if verbose:
                print('writing file: {}'.format(figfile))
            fig.savefig(figfile, dpi=dpi)

    if show and not print_figs:
        if show == 'draw':
            if use_bokeh:
                warn('draw not implemented with bokeh', RuntimeWarning)
            else:
                mplt.draw()
        elif show is not False:
            if use_bokeh:
                for fig in figs:
                    mplt.show(fig)
            else:
                mplt.show()
    elif print_figs and not use_bokeh:
        for fig in figs:
            mplt.close(fig)
    else:
        return figs


infer = infer1
map_plot = map_plot1




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
import polytope as p
from tramway.tessellation import Voronoi, Partition
from scipy.spatial import ConvexHull


setup = {'name':        'srtesseler.density',
         'provides':    'density',
         'input_type':  'Partition'}


def convex_hull(vertices=None, vertex_indices=None, points=None, point_assigned=None,
        partition=None, cell_index=None):
    # recover the vertices of the Voronoi cell
    if vertex_indices is None:
        vertex_indices = partition.tessellation.cell_vertices[cell_index]
        if vertices is None:
            vertices = partition.tessellation.vertices
    if np.any(vertex_indices < 0):
        valid_indices = vertex_indices[0<=vertex_indices]
        vertices = vertices[valid_indices]
        # for cells with missing vertices (to the infinite),
        # consider some inner points to possibly push the hull outward
        if point_assigned is None:
            point_assigned = partition.cell_index == cell_index
        if points is None:
            points = partition.points
        assigned_points = points[point_assigned]
        try:
            hull = ConvexHull(np.r_[vertices, assigned_points])
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            hull = None
    else:
        hull = ConvexHull(vertices[vertex_indices])
    return hull


def infer_srtesseler_density(cells, volume_weighted=True, **kwargs):
    """
    adapted from:

        SR-Tesseler: a method to segment and quantify localization-based super-resolution microscopy data.
        Levet, F., Hosy, E., Kechkar, A., Butler, C., Beghin, A., Choquet, C., Sibarita, J.B.
        Nature Methods 2015; 12 (11); 1065-1071.

    """
    points = cells.locations
    is_densest_tessellation = isinstance(cells.tessellation, Voronoi) and cells.number_of_cells == len(points)
    if is_densest_tessellation:
        tessellation = cells.tessellation
        partition = cells
    else:
        tessellation = Voronoi()
        tessellation.tessellate(points[['x','y']])
        partition = Partition(points, tessellation)
    #areas = tessellation.cell_volume # the default implementation for cell_volume
                                      # generates a non-infinite estimation for opened cells
    areas = np.full(tessellation.number_of_cells, np.nan)
    indices, densities, unit_polygons = [], [], []
    for i, surface_area in enumerate(areas):
        unknown_area = np.isnan(surface_area) or np.isinf(surface_area) or surface_area == 0
        polygons_required = not is_densest_tessellation and volume_weighted
        if unknown_area or polygons_required:
            hull = convex_hull(partition=partition, cell_index=i)
        if unknown_area:
            if hull is None:
                continue
            density = 1./ hull.volume # should also divide by the number of frames
        else:
            density = 1./ surface_area
        indices.append(i)
        densities.append(density)
        if polygons_required:
            unit_polygons.append(p.Polytope(hull.equations[:,[0,1]], -hull.equations[:,2]))
    unit_density = pd.Series(index=indices, data=densities)
    if is_densest_tessellation:
        density = unit_density
    else:
        indices, densities = [], []
        for i in range(cells.number_of_cells):
            assigned = cells.cell_index == i
            if not np.any(assigned):
                indices.append(i)
                densities.append(0.)
                continue
            local_polygons, = np.nonzero(assigned)
            assert 0 < local_polygons.size
            if volume_weighted:
                hull = convex_hull(partition=cells, cell_index=i, point_assigned=assigned)
                if hull is None:
                    continue
                polygon = p.Polytope(hull.equations[:,[0,1]], -hull.equations[:,2])
                weighted_sum = 0.
                total_area = 0.
                for j in local_polygons:
                    intersection = p.intersect(unit_polygons[j], polygon)
                    total_area += intersection.volume
                    weighted_sum += intersection.volume * unit_density[j]
                average_density = weighted_sum / total_area
            else:
                average_density = unit_density[local_polygons].mean()
            indices.append(i)
            densities.append(average_density)
        density = pd.Series(index=indices, data=densities)
    return pd.DataFrame(dict(density=density))


__all__ = ['infer_srtesseler_density', 'setup']





"""
Standard analysis steps for RWAnalyzer.pipeline
"""

from ..roi import FullRegion
from ..spt_data import _RWAFile, StandaloneDataItem, RWAFiles
from . import PipelineStage
from ..attribute import single
from ..artefact import commit_as_analysis
import os.path
import numpy as np
from tramway.core import load_xyt
from collections import defaultdict
import logging

module_logger = logger = logging.getLogger(__name__)


def _spt_source_name(f):
    try:
        alias = f.alias
    except AttributeError:
        alias = None
    return f.source if alias is None else alias

def _filters(kwargs):
    asr_filters = {}
    for k in kwargs:
        if k.startswith('spt_') or k.startswith('roi_'):
            asr_filters[k[4:]] = kwargs[k]
    return asr_filters

def _placeholder_for_root_data(tree):
    tree._data = None


def tessellate(label=None, roi_expected=False, spt_data=True, tessellation='freeze', **kwargs):
    """
    Returns a standard pipeline stage for SPT data sampling.

    Although the name 'tessellate' refers merely to the spatial
    segmentation, time segmentation is also handled, if defined.
    The name alludes to the standalone
    :func:`tramway.helper.tessellation.tessellate` function.

    The default granulariy is set to 'spt data',
    whereas in principle the finest valid granularity is 'roi'.
    The tessellation step is fast enough to a avoid the overhead
    of saving as many files as regions of interest.

    .. important::

        If the input SPT data is defined as ascii file(s),
        the corresponding *.rwa* files will be overwritten.

        If analysis trees already reference analysis artefacts
        for the specified label, these artefacts are also
        overwritten.

    This stage building function features two mechanisms to
    make the resulting *.rwa* files smaller in size:

    * with option `spt_data='placeholder'`, the SPT data can be
      omitted in the *.rwa* files;
      see also :func:`restore_spt_data`;
    * the spatial tessellation can be freezed (default);
      this implies the tesellation cannot be updated any longer
      with extra data.

    """

    asr_filters = _filters(kwargs)

    def _tessellate(self):

        dry_run = True

        for f in self.spt_data:

            # for logging
            source_name = _spt_source_name(f)
            # for data formatting
            any_full_region = False

            with f.autosaving() as tree:

                for r in f.roi.as_support_regions(**asr_filters):

                    if isinstance(r, FullRegion):
                        if roi_expected:
                            continue
                        any_full_region = True
                        msg = f"tessellating source: '{source_name}'..."
                    else:
                        roi_label = r.label
                        msg = f"tessellating roi: '{roi_label}' (in source '{source_name}')..."

                    # get the SPT data
                    df = r.crop()

                    # filter some translocations out
                    df = r.discard_static_trajectories(df)

                    # tessellate
                    logger.info(msg)
                    sampling = self.sampler.sample(df)

                    dry_run = False

                    # store
                    if tessellation=='freeze':
                        try:
                            sampling.tessellation.freeze()
                        except AttributeError:
                            pass
                    r.add_sampling(sampling, label=label)

                if spt_data == 'placeholder':
                    ## [does not work:] make the dataframe empty and keep the additional attributes
                    #df = f.dataframe
                    #f._dataframe = df[np.zeros(df.shape[0], dtype=bool)]
                    tree.hooks.append(_placeholder_for_root_data)
                elif not any_full_region:
                    # save the full dataset with reasonnable precision to save storage space
                    f.set_precision('single')
                    # the data in the roi is stored separately and will keep the original precision

        if dry_run:
            logger.info('stage skipped')
            diagnose(self)

    return PipelineStage(_tessellate, granularity='spt data')


def infer(map_label=None, sampling_label=None, roi_expected=False, overwrite=False,
        single_path=False, **kwargs):
    """
    Returns a standard pipeline stage for inferring model parameters on each region of interest,
    or SPT data item if no roi are defined.

    The default granularity is set to 'roi', which is suitable for computer-intensive
    inferences such as 'DV'.

    With default ``overwrite=False``, if the specified output label `map_label` already
    exists in analysis trees, the inference is skipped for these analysis trees and the
    corresponding artefact not overwritten.
    However, if `map_label` is :const:`None`, `overwrite` is ignored and the stage acts like if
    `overwrite` were :const:`True`.
    """

    asr_filters = _filters(kwargs)

    def _infer(self):

        dry_run = True

        for r in self.roi.as_support_regions(**asr_filters):

            # get the input data
            sampling = r.get_sampling(sampling_label)

            if not overwrite and map_label in sampling.subtree:
                # skip the already-processed data
                continue

            source_name = _spt_source_name(r._spt_data)
            if isinstance(r, FullRegion):
                if roi_expected:
                    continue
                msg = f"inferring on source: '{source_name}'..."
            else:
                roi_label = r.label
                msg = f"inferring on roi: '{roi_label}' (in source '{source_name}')..."

            with r.autosaving() as tree:

                logger.info(msg)
                maps = self.mapper.infer(sampling)

                dry_run = False

                # store
                maps.commit_as_analysis(map_label)

                if single_path:
                    for label in tree:
                        if label != sampling_label:
                            del tree[label]
                    for label in tree[sampling_label]:
                        if label != map_label:
                            del tree[sampling_label][label]

        if dry_run:
            logger.info('stage skipped')
            diagnose(self)

    return PipelineStage(_infer, granularity='roi')


def reload(skip_missing=True):
    """
    Returns a pipeline stage to reload the generated *.rwa* files.

    This is useful if a first computation - *e.g.* `tessellate` - was dispatched
    (including with the `environments.LocalHost` environment)
    and the *.rwa* files were generated and retrieved from the worker
    environments.
    In the case a second stage has to be dispatched - *e.g.* `infer`,
    the local pipeline must be updated with the content of these retrieved files.

    This may also be required on the worker side for the next stages,
    for example if the SPT data was initially defined from ascii files
    and the coming stages should take over from the corresponding *.rwa* files.
    As a consequence, this `reload` stage is set with `requires_mutability=True`.

    """

    def _reload(self):
        self.spt_data.reload_from_rwa_files(skip_missing=skip_missing)

    return PipelineStage(_reload, requires_mutability=True)

def restore_spt_data():
    """
    Reintroduces the original SPT dataframe into the *.rwa* files.

    This post-processing stage - typically post-infer stage - is expected to be
    called only when the `tessellate` stage was called before with option
    `spt_data='placeholder'`.

    This stage was tested with SPT data defined as :class:`~..spt_data.SPTAsciiFiles`
    or :class:`..spt_data.RWAFiles`.
    However, with :class:`SPTAsciiFiles`, the `alias` attribute should be defined.

    """
    from ..spt_data import RWAFile

    def _restore(self):
        for f in self.spt_data:
            if not isinstance(f, RWAFile):
                try:
                    alias = f.alias
                except AttributeError:
                    alias = None
                if alias is None:
                    raise NotImplementedError('alias is not defined')
                #
                analyzer = self._eldest_parent
                analyzer.spt_data.reload_from_rwa_files()
                f = single(analyzer.spt_data.filter_by_source(alias))
            #
            spt_ascii_file = f.analyses.metadata['datafile']
            rwa_file = f.filepath
            local_spt_file = os.path.join(os.path.dirname(rwa_file), os.path.basename(spt_ascii_file))
            if os.path.isfile(local_spt_file):
                df = load_xyt(local_spt_file)
                restore = no_dataframe = f.dataframe is None
                if not no_dataframe:
                    cols, cols_ = f.dataframe.columns, df.columns
                    restore = len(cols)==len(cols_) and all([ c==c_ for c, c_ in zip(cols, cols_) ])
                if restore:
                    with f.autosaving() as tree:
                        f._dataframe = df
                        tree.flag_as_modified()
                else:
                    logger.error('column names do not match with file: {}'.format(local_spt_file))
            else:
                logger.error('could not find file: {}'.format(local_spt_file))

    return PipelineStage(_restore, granularity='spt data')


def tessellate_and_infer(map_label=None, sampling_label=None, spt_data=True, overwrite=False,
        roi_expected=False, load_rwa_files=None, **kwargs):
    """
    Combines `tessellate` and `infer` at roi granularity.

    *new in 0.6*: unless ``overwrite=True`` or ``load_rwa_files=False``,
    *.rwa* files are loaded if the SPT data source has not already
    been loaded from such files.
    """

    save_active_branches_only = not (overwrite or map_label is None)

    if overwrite:
        if load_rwa_files is None:
            load_rwa_files = False
        elif load_rwa_files is True:
            raise ValueError('overwrite and load_rwa_files are both True')

    def _tessellate_and_infer(self):

        if self._eldest_parent.env.initialized:
            _save_active_branches_only = save_active_branches_only
        else:
            _save_active_branches_only = False

        dry_run = True

        # added in 0.6.1; useful only without environment?
        if not (load_rwa_files is False or \
                isinstance(self.spt_data, (StandaloneDataItem, RWAFiles))):
            _load_rwa_files = load_rwa_files
            if load_rwa_files is None:
                for f in self.spt_data:
                    if f.source:
                        rwa_file = os.path.splitext(f.source)[0]+'.rwa'
                        _load_rwa_files = os.path.isfile(rwa_file)
                    break
            if _load_rwa_files:
                self.spt_data.reload_from_rwa_files(skip_missing=True)

        for f in self.spt_data:

            source_dry_run = True

            # for logging
            source_name = _spt_source_name(f)
            # for data formatting
            any_full_region = False

            if _save_active_branches_only:
                active_labels = defaultdict(set)

            # new in 0.6; actually works only with StandaloneDataItem
            if not (load_rwa_files is False or isinstance(f, _RWAFile)):
                try:
                    rwa_file = os.path.splitext(f.source)[0]+'.rwa'
                except Exception: # Exception excludes KeyboardInterrupt and SystemExit
                    pass
                else:
                    if os.path.isfile(os.path.expanduser(rwa_file)):
                        logger.info(f"loading .rwa file for source: {source_name}...")
                        try:
                            f = f.reload_from_rwa_files()
                        except AttributeError as e:
                            if e.args[0].endswith("'reload_from_rwa_files'"):
                                logger.warning('existing .rwa file found; none of `overwrite` or `load_rwa_files` is True')
                            raise

            with f.autosaving() as tree:

                for r in f.roi.as_support_regions():

                    if isinstance(r, FullRegion):
                        if roi_expected:
                            continue
                        any_full_region = True
                        msg = f"{{}} source: '{source_name}'..."
                    else:
                        roi_label = r.label
                        msg = f"{{}} roi: '{roi_label}' (in source '{source_name}')..."
                    def log(op):
                        logger.info(msg.format(op))

                    if sampling_label is None:
                        label = r.label
                    elif callable(sampling_label):
                        label = sampling_label(r.label)
                    else:
                        if not source_dry_run:
                            logger.warning("multiple roi bound to single sampling label; make sampling_label callable")
                        label = sampling_label

                    # control predicates
                    _tessellate = overwrite or label is None or label not in tree.labels
                    _infer = _tessellate or overwrite or map_label is None or \
                            map_label not in tree[label].labels

                    if _tessellate:

                        # get the SPT data
                        df = r.crop()

                        if df.empty:
                            raise ValueError(f'not a single displacement in roi {roi_label}')

                        # filter some translocations out
                        df = r.discard_static_trajectories(df)

                        if df.empty:
                            raise ValueError(f'all the trajectories are static in roi {roi_label}')

                        # tessellate
                        log('tessellating')
                        sampling = self.sampler.sample(df)

                    elif _infer:
                        sampling = r.get_sampling(sampling_label).data

                    if _infer:

                        # infer
                        log('inferring')
                        maps = self.mapper.infer(sampling)

                        dry_run = source_dry_run = False

                    # store
                    if _tessellate:
                        # ... the tessellation
                        try:
                            sampling.tessellation.freeze()
                        except AttributeError:
                            pass
                        r.add_metadata(sampling, inplace=True)
                        sampling = commit_as_analysis(label, sampling, parent=r)
                        #
                    elif _infer:
                        sampling = r.get_sampling(sampling_label)
                    if _infer:
                        # ... and the inferred maps
                        maps     = commit_as_analysis(map_label, maps, parent=sampling)
                        #
                        assert tree.modified(recursive=True)
                        if _save_active_branches_only:
                            active_labels[label].add(map_label)

                if _save_active_branches_only:
                    for label0 in list(tree.labels):
                        if label0 in active_labels:
                            for label1 in list(tree[label0].labels):
                                if label1 not in active_labels[label0]:
                                    tree[label0].comments
                                    del tree[label0][label1]
                        else:
                            tree.comments
                            del tree[label0]

                if spt_data=='placeholder':
                    tree.hooks.append(_placeholder_for_root_data)

        if dry_run:
            logger.info('stage skipped')
            diagnose(self)

    return PipelineStage(_tessellate_and_infer,
            granularity='roi',
            update_existing_rwa_files=not overwrite)


def diagnose(self):
    """
    Checks for filter consistency.

    To be called only in the case of proven dry runs.
    """
    from ..spt_data import _normalize

    try:
        sources = self.env.selectors['source']
    except AttributeError:
        sources = ()
    if isinstance(sources, str):
        sources = (sources,)

    try:
        alias = all([ bool(f.alias) for f in self._eldest_parent.spt_data ])
    except AttributeError:
        alias = False
    if alias:
        requested = set(sources)
        available = set(self.spt_data.aliases)
    else:
        requested = set([ _normalize(_s) for _s in sources ])
        available = set([ _normalize(_f.source) for _f in self.spt_data ])

    if bool( requested - available ):
        logger.critical(f"""
The local worker was assigned the following SPT data source(s):
   {requested}
but listed other sources:
   {available}
This is known to happen when files are listed from the file system
and a subset of the listed files is selected based on their order
in the list.
Avoid doing so!
Filename ordering varies from a file system to another.
""")
        raise RuntimeError


__all__ = ['tessellate', 'infer', 'reload', 'restore_spt_data', 'tessellate_and_infer']




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This package is no longer maintained and is superseded by the *mpl*
functionalities of :class:`~tramway.analyzer.RWAnalyzer`.

See for example :meth:`~tramway.analyzer.images._RawImage.to_color_movie`,
:meth:`~tramway.analyzer.spt_data.mpl.Mpl.animate` and
:meth:`~tramway.analyzer.mapper.mpl.Mpl.animate`.
"""

try:
    input = raw_input # Py2
except NameError:
    pass

import os
#import matplotlib
#matplotlib.use('Agg')
from matplotlib.animation import FFMpegWriter

try:
    import cv2
except:
    try: # Windows
        os.startfile
    except:
        class VideoReader(object):
            def __init__(self, *args, **kwargs):
                import cv2 # cannot read the movie
    else:
        class VideoReader(object):
            def __init__(self, filepath, *args, **kwargs):
                self.path = filepath
            def play(self, **kwargs):
                os.startfile(self.path)
else:
    # mostly borrowed from:
    # https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_gui/py_video_display/py_video_display.html#playing-video-from-file
    class VideoReader(object):
        def __init__(self, filepath, *args, **kwargs):
            self.path = filepath
            self.cap = cv2.VideoCapture(self.path)
        def play(self, fps=20, **kwargs):
            frame_duration = int(1000./float(fps))
            if not self.cap.isOpened():
                self.cap.open(self.path)
                if not self.cap.isOpened():
                    raise RuntimeError('cannot open the movie')
            while self.cap.isOpened():
                ok, frame = self.cap.read()
                if not ok: break
                cv2.imshow('frame', frame)
                # wait and listen key events
                key = cv2.waitKey(frame_duration)
                if key & 0xFF == ord('q'): # quit on 'q'
                    break
                elif key == 27: # quit on ESC
                    break
                # check that window is still opened
                if cv2.getWindowProperty('frame', cv2.WND_PROP_VISIBLE) < 1:
                    # in old opencv-python versions, use cv2.WND_PROP_AUTOSIZE instead
                    break
            self.cap.release()
            cv2.destroyAllWindows()


class Aborted(Exception):
    pass


class VideoWriter(object):
    def __init__(self, frame_rate, bit_rate=None, dots_per_inch=200, figure=None, axes=None,
            axis=True, verbose=True, **kwargs):
        self.fps = kwargs.get('fps', None)
        if self.fps is None:
            self.fps = frame_rate
        kwargs['fps'] = self.fps
        if 'bitrate' not in kwargs:
            kwargs['bitrate'] = bit_rate
        self.grab = FFMpegWriter(**{kw: arg for kw, arg in kwargs.items() if arg is not None})

        self.finally_close = False
        if figure is None:
            if axes is None:
                # always import pyplot as late as possible to allow for backend selection
                import matplotlib.pyplot as plt
                figure, axes = plt.subplots()
                figure.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
                self.finally_close = True
            else:
                figure = axes.get_figure()
        elif axes is None:
            axes = figure.gca()
        if axis in (False, 'off'):
            axes.set_axis_off()
        self.dpi = dots_per_inch
        self.figure = figure
        self.axes = axes
        self.verbose = verbose

    def saving(self, output_file):
        output_file = os.path.expanduser(output_file)
        if self.verbose and os.path.exists(output_file) and 0 < os.path.getsize(output_file):
            answer = input("overwrite file '{}': [N/y] ".format(output_file))
            if not (answer and answer[0].lower() == 'y'):
                raise Aborted
        return self.grab.saving(self.figure, output_file, self.dpi)

    def range(self, N):
        trange = range
        if self.verbose:
            try:
                import tqdm
            except ImportError:
                pass
            else:
                trange = tqdm.trange
        return trange(int(N))

    def grab_frame(self, *args, **kwargs):
        self.grab.grab_frame(*args, **kwargs)


class VideoWriterReader(object):
    def __init__(self, filepath, *args, **kwargs):
        self.path = filepath
        self.unlink_on_exit = False
        self.writer = VideoWriter(*args, **kwargs)

    @property
    def fps(self):
        return self.writer.fps
    @fps.setter
    def fps(self, f):
        self.writer.fps = f

    @property
    def figure(self):
        return self.writer.figure
    @figure.setter
    def figure(self, f):
        self.writer.figure = f

    @property
    def axes(self):
        return self.writer.axes
    @axes.setter
    def axes(self, a):
        self.writer.axes = a

    @property
    def verbose(self):
        return self.writer.verbose
    @verbose.setter
    def verbose(self, v):
        self.writer.verbose = v

    def __enter__(self):
        if self.path is None:
            import tempfile
            fd, self.path = tempfile.mkstemp(suffix='.mp4')
            os.close(fd)
            self.unlink_on_exit = True
        return self

    def __exit__(self, *args):
        if self.writer.finally_close:
            # ideally this should happen in VideoWriter
            import matplotlib.pyplot as plt
            plt.close(self.figure)
        if self.unlink_on_exit:
            os.unlink(self.path)

    def saving(self):
        return self.writer.saving(self.path)

    def play(self, fps=None, **kwargs):
        VideoReader(self.path).play(fps=self.fps if fps is None else fps)

    def range(self, N):
        return self.writer.range(N)

    def grab_frame(self, *args, **kwargs):
        return self.writer.grab_frame(*args, **kwargs)


__all__ = [ 'VideoReader', 'VideoWriter', 'VideoWriterReader', 'Aborted' ]




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import *
from tramway.tessellation.base import *
import numpy as np
import pandas as pd
import copy
import scipy.sparse as sparse


class NestedTessellations(Tessellation):
    """Tessellation of tessellations.

    When nesting, the parent tessellation should have been grown already.

    `tessellation` grows all the nested tessellations.

    In `__init__`, `tessellation` and `cell_index`,
    the `scaler` (if any), `args` (if any) and `kwargs` arguments
    only apply to the nested tessellations.
    """
    __slots__ = ('_parent', '_children', 'child_factory', \
        'parent_index_arguments', 'child_factory_arguments', \
        '_cell_centers')

    __lazy__ = Tessellation.__lazy__ + ('cell_label', 'cell_adjacency', 'adjacency_label', \
        'cell_centers')

    def __init__(self, scaler=None, parent=None, factory=None,
            parent_index_arguments={}, **kwargs):
        Tessellation.__init__(self, scaler)
        self._parent = parent
        self._children = {}
        self.child_factory = factory
        self.parent_index_arguments = dict(parent_index_arguments)
        self.child_factory_arguments = kwargs
        self._cell_centers = None

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, tessellation):
        if tessellation != self._parent:
            self.children = {}
            self._parent = tessellation

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, tessellations):
        self.cell_label = None
        self.cell_adjacency = None
        self.adjacency_label = None
        self.cell_centers = None
        self._children = tessellations

    def _parent_index(self, points):
        if isinstance(self.parent, Partition):
            parent = self.parent
        else: # preferred usage
            parent = Partition(tessellation=self.parent)
        if points is not parent.points or \
                (self.parent_index_arguments and parent._cell_index is None):
            parent.points = points
            parent.cell_index = parent.tessellation.cell_index(points, \
                    **self.parent_index_arguments)
        parent_index = format_cell_index(parent.cell_index, 'pair') # ignore association weights
        pt_ids, cell_ids = parent_index
        if isinstance(points, pd.DataFrame):
            def rows(pts, ids):
                return pts.iloc[ids,:]
        else:
            def rows(pts, ids):
                return pts[ids]
        return (pt_ids, cell_ids, rows)

    def tessellate(self, points, *args, **kwargs):
        # initialize `self.scaler`;
        # if we didn't, we should pass copies of `self.scaler` instead, to `self.child_factory`
        any_child = self.child_factory(scaler=self.scaler, **self.child_factory_arguments)
        any_child._preprocess(points)
        #
        pt_ids, cell_ids, rows = self._parent_index(points)
        self.children = {}
        for u in np.unique(cell_ids):
            child = self.child_factory(scaler=self.scaler, **self.child_factory_arguments)
            child_pts = rows(points, pt_ids[cell_ids==u])
            if child_pts.size:
                child.tessellate(child_pts, *args, **kwargs)
                self.children[u] = child

    def cell_index(self, points, *args, **kwargs):
        point_count = points.shape[0]
        #if isinstance(points, pd.DataFrame):
        #       point_count = max(point_count, points.index.max()+1) # NO!
        # point indices are row indices and NOT rows labels
        parent_pt_ids, parent_cell_ids, rows = self._parent_index(points)
        _is_array_ = _is_pair_ = _is_sparse_ = False # (exclusive) type flags
        _first_ = True # initialization flag
        _type_error_ = TypeError('multiple nested partition types; `format` should be enforced')
        pt_cell, pt_ids, cell_ids = [], [], []
        cell_count = 0
        for u in self.children:
            child_cell_count = self.children[u].cell_adjacency.shape[0]
            child_pt_ids = parent_pt_ids[parent_cell_ids==u]
            if child_pt_ids.size: # if cell is empty
                child_partition = self.children[u].cell_index(
                    rows(points, child_pt_ids), *args, **kwargs)
                # TODO: test cases such that not any cell-point association is found;
                # currently, ``continue`` is not an option but
                # _first_ is made False anyway
                if sparse.issparse(child_partition):
                    if _first_:
                        _is_sparse_ = True
                        if args:
                            _format = args.pop(0)
                        elif 'format' in kwargs:
                            _format = kwargs['format']
                        else:
                            _format = None
                        kwargs['format'] = 'coo' # for the next calls to `cell_index`
                        if not sparse.isspmatrix_coo(child_partition):
                            child_partition = child_partition.tocoo()
                    elif not _is_sparse_:
                        raise _type_error_
                    # use `pt_ids` as `rows`, `cell_ids` as `cols` and `pt_cell` as `data`
                    _pt_cell = child_partition.data
                    _pt_ids = child_pt_ids[child_partition.rows]
                    _cell_ids = cell_count + child_partition.cols
                    pt_cell.append(_pt_cell)
                    pt_ids.append(_pt_ids)
                    cell_ids.append(_cell_ids)
                elif isinstance(child_partition, np.ndarray):
                    if cell_count:
                        child_partition[0<=child_partition] += cell_count
                    if _first_:
                        _is_array_ = True
                        pt_cell = np.full(point_count, -1,
                            dtype=child_partition.dtype)
                    else:
                        if not _is_array_:
                            raise _type_error_
                        offset = 0
                    pt_cell[child_pt_ids] = child_partition
                elif isinstance(child_partition, tuple):
                    _pt_ids, _cell_ids = child_partition
                    _pt_ids = child_pt_ids[_pt_ids]
                    _cell_ids += cell_count
                    if _first_:
                        _is_pair_ = True
                    elif not _is_pair_:
                        raise _type_error_
                    pt_ids.append(_pt_ids)
                    cell_ids.append(_cell_ids)
                else:
                    raise ValueError('partition type not supported')
                _first_ = False
            cell_count += child_cell_count
        if isinstance(pt_cell, list) and pt_cell:
            pt_cell = np.concatenate(pt_cell)
        if pt_ids:
            pt_ids = np.concatenate(pt_ids)
        if cell_ids:
            cell_ids = np.concatenate(cell_ids)
        if _is_sparse_:
            partition = sparse.coo_matrix((pt_cell, (pt_ids, cell_ids)),
                shape=(point_count, cell_count))
            if _format is None or _format in ['matrix', 'coo']:
                return partition
            else:
                return dict(
                        csc=partition.tocsc,
                        csr=partition.tocsr,
                        dia=partition.todia,
                        dok=partition.todok,
                        lil=partition.tolil,
                    )[_format]()
        elif _is_array_:
            return pt_cell
        elif _is_pair_:
            return (pt_ids, cell_ids)

    # cell_adjacency property
    @property
    def cell_adjacency(self):
        if self._cell_adjacency is None:
            _all_label_ = True
            _any_label_ = False
            labels = set()
            for child in self.children.values():
                if child.adjacency_label is None:
                    _all_label_ = False
                else:
                    _any_label_ = True
                    labels |= set(list(child.adjacency_label))
            if labels:
                if not _all_label_:
                    na = min(labels) - 1
                    labels.add(na)
                labels = list(labels)
                self._adjacency_label = np.array(labels)
                labels = dict(zip(labels, range(len(labels))))
            row, col, data = [], [], []
            cell_count = 0
            for u in self.children:
                A = self.children[u].cell_adjacency.tocoo()
                A.row += cell_count
                A.col += cell_count
                if self.children[u].adjacency_label is None:
                    if labels:
                        A.data = np.zeros_like(A.data, dtype=int)
                else:
                    A.data = np.array([ labels[e] for e in A.data ])
                row.append(A.row)
                col.append(A.col)
                data.append(A.data)
                cell_count += A.shape[0]
            row = np.concatenate(row)
            col = np.concatenate(col)
            data = np.concatenate(data)
            self._cell_adjacency = sparse.coo_matrix((data, (row, col)),
                shape=(cell_count, cell_count))
        return self.__returnlazy__('cell_adjacency', self._cell_adjacency)

    @cell_adjacency.setter
    def cell_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    # cell_label
    @property
    def cell_label(self):
        if self._cell_label is None:
            _err_ = AttributeError('not all the nested tessellations have cell labels')
            _missing_label_ = False
            labels = []
            for u in self.children:
                child_labels = self.children[u].cell_label
                if child_labels:
                    if _missing_label_:
                        raise _err_
                    labels.append(child_labels)
                else:
                    if labels:
                        raise _err_
                    _missing_label_ = True
            if _missing_label_:
                self._cell_label = False
            else:
                self._cell_label = np.concatenate(labels)
        if self._cell_label is False:
            return None
        else:
            return self.__returnlazy__('cell_label', self._cell_label)

    @cell_label.setter
    def cell_label(self, label):
        self.__lazysetter__(label)

    # adjacency_label
    @property
    def adjacency_label(self):
        if self._adjacency_label is None:
            # this doesn't cost anything if `_cell_adjacency` has already been built
            self.cell_adjacency # makes `_adjacency_labels`
        return self.__returnlazy__('adjacency_label', self._adjacency_label)

    @adjacency_label.setter
    def adjacency_label(self, label):
        self.__lazysetter__(label)

    def freeze(self):
        for child in self.children.values():
            child.freeze()

    @property
    def cell_centers(self):
        if self._cell_centers is None:
            try:
                centers = []
                for child in self.children.values():
                    centers.append(child.cell_centers)
            except AttributeError:
                raise AttributeError("'NestedTessellations' object has no attribute 'cell_centers'")
            else:
                # note that a dimension may not represent the same variable
                # from a child to another
                self._cell_centers = np.vstack(centers)
        return self.__returnlazy__('cell_centers', self._cell_centers)

    @cell_centers.setter
    def cell_centers(self, centers):
        self.__lazyassert__(centers)

    def child_cell_indices(self, u):
        """
        Arguments:

            u (int): child index.

        Returns:

            slice: child's cell indices.
        """
        cell_count = 0
        for v in self.children:
            child = self.children[v]
            child_cell_count = child.cell_adjacency.shape[0]
            if u == v:
                return slice(cell_count, cell_count + child_cell_count)
            cell_count += child_cell_count
        raise IndexError('no such child index: {}'.format(u))



__all__ = ['NestedTessellations']





try:
    from .lazy import PermissionError
except ImportError:
    pass

class NaNWarning(RuntimeWarning):
    pass

class EfficiencyWarning(RuntimeWarning):
    pass

class MissingSupportWarning(UserWarning):
    pass

class IOWarning(Warning):
    pass

class FileNotFoundWarning(IOWarning):
    pass

class MultipleArgumentError(ValueError):
    def __str__(self):
        assert self.args and self.args[1:] and all( isinstance(a, str) for a in self.args )
        args = [ "'{}'".format(a) for a in self.args ]
        s = ' or '.join((', '.join(args[:-1]), args[-1]))
        return 'please define either ' + s

class MisplacedAttributeWarning(UserWarning):
    pass

class SideEffectWarning(UserWarning):
    pass

class RWAFileException(IOError):
    def __init__(self, filepath=None, exc=None):
        self.filepath = filepath
        self.exc = exc
    def __str__(self):
        if self.filepath is None:
            return 'cannot find any analysis tree'
        else:
            return 'cannot find any analysis tree in file: '+self.filepath




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import AnalyzerNode
import numpy as np
import pandas as pd


class Mpl(AnalyzerNode):
    """
    Matplotlib plotting utilities for 2D data; no time support.
    """
    __slots__ = ()
    def plot(self, x, *args, origin=None, frame=None, **kwargs):
        """
        Converts localization data from micrometers to pixels and plots them.

        Can overlay locations onto a frame image as plotted with `imshow` and
        default parameters (``origin='upper'``).
        """
        if isinstance(x, pd.DataFrame):
            x, y = x['x'], x['y']
        else:
            y = args[0]
            args = args[1:]
        try:
            axes = kwargs.pop('axes')
        except KeyError:
            import matplotlib.pyplot as axes
        img = self._parent
        if frame is None:
            height = img.height
        else:
            #axes.imshow(frame)
            height = frame.shape[0]
        x, y = np.array(x), np.array(y)
        if origin is not None:
            x -= origin[0]
            y -= origin[1]
        x /= img.pixel_size
        y /= img.pixel_size
        if origin is None and img.loc_offset is not None:
            x -= img.loc_offset[0]
            y -= img.loc_offset[1]
        return axes.plot(x, height-1-y, *args, **kwargs)

    def roi(self, frame, origin, *args, axes=None, colormap='gray', **kwargs):
        """
        Plots a frame corresponding to a region of interest.

        `origin` should be provided by :meth:`cropping_bounds`,
        if `frame` is given by :meth:`crop_frames`.

        Extra input arguments are passed to the :func:`imshow` function.

        This method works in combination with :meth:`spt_data.mpl.Mpl.plot`.
        """
        m, n = frame.shape
        px = self._parent.pixel_size
        x0, y0 = origin[0] - .5 * px, origin[1] - .5 * px
        x1, y1 = origin[0] + (m - .5) * px, origin[1] + (n - .5) * px
        try:
            colormap = kwargs.pop('cmap')
        except KeyError:
            pass
        for arg in ('origin', 'extent'):
            try:
                kwargs.pop(arg)
            except KeyError:
                pass
            else:
                self.logger.warning(f"argument '{arg}' ignored")
        if axes is None:
            import matplotlib.pyplot as axes
        return axes.imshow(frame, cmap=colormap, extent=[x0, x1, y0, y1],
                **kwargs)


__all__ = ['Mpl']

